@@ -1,5 +1,5 @@
 //  crm_correlate.c  - Controllable Regex Mutilator,  version v1.0
-//  Copyright 2001-2004  William S. Yerazunis, all rights reserved.
+//  Copyright 2001-2006  William S. Yerazunis, all rights reserved.
 //  
 //  This software is licensed to the public under the Free Software
 //  Foundation's GNU GPL, version 2.  You may obtain a copy of the
@@ -39,7 +39,8 @@ extern char *tempbuf;
 //    learned to the target file.  We don't care about the /regexes/
 //
 
-int crm_expr_correlate_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
+int crm_expr_correlate_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb, 
+			      char *txtptr, long txtstart, long txtlen)
 {
   //     learn the given text as correlative text
   //     belonging to a particular type.
@@ -97,7 +98,7 @@ int crm_expr_correlate_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   sense = +1;
   if (apb->sflags & CRM_NOCASE)
     {
-      cflags = cflags || REG_ICASE;
+      cflags = cflags | REG_ICASE;
       eflags = 1;
       if (user_trace)
 	fprintf (stderr, "turning oncase-insensitive match\n");
@@ -139,14 +140,24 @@ int crm_expr_correlate_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
       //      file didn't exist... create it
       FILE *f;
       if (user_trace)
-	fprintf (stderr, "\nHad to create new correlate file %s\n", &htext[i]);
+	{
+	  fprintf (stderr, "\nCreating new correlate file %s\n", &htext[i]);
+	  fprintf (stderr, "Opening file %s for write\n", &htext[1]);
+	};
       f = fopen (&htext[i], "wb");
       if (!f)
 	{
 	  fprintf (stderr, 
 		"\n Couldn't open your new correlate file %s for writing; errno=%d .\n",
 		 &htext[i], errno);
-	  exit (EXIT_FAILURE);
+
+	  if (engine_exit_base != 0) 
+	    {
+	      exit (engine_exit_base + 1);
+	    }
+	  else
+	    exit (EXIT_FAILURE);
+
         };
       //      fputc ('\001', f); don't do any output at all.
       made_new_file = 1;
@@ -167,8 +178,8 @@ int crm_expr_correlate_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   hfd = open (&(htext[i]), O_RDWR);
   if (hfd < 0) 
     {
-      fev = fatalerror ("Couldn't open the correlation file named: ",
-			&htext[i]);
+      fev = fatalerror5 ("Couldn't open the correlation file named: ",
+			 &htext[i], CRM_ENGINE_HERE);
       return (fev);
     };
 
@@ -191,8 +202,8 @@ int crm_expr_correlate_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   if (vht[vhtindex] == NULL)
     {
       long q;
-      q = fatalerror (" Attempt to LEARN from a nonexistent variable ",
-		  ltext);
+      q = fatalerror5 (" Attempt to LEARN from a nonexistent variable ",
+		       ltext, CRM_ENGINE_HERE);
       return (q);
     };
   mdw = NULL;
@@ -203,7 +214,8 @@ int crm_expr_correlate_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   if (mdw == NULL)
     {
       long q;
-      q = fatalerror (" Bogus text block containing variable ", ltext);  
+      q = fatalerror5 (" Bogus text block containing variable ", 
+		       ltext, CRM_ENGINE_HERE);  
       return (q);
     }
   textoffset = vht[vhtindex]->vstart;
@@ -229,7 +241,8 @@ int crm_expr_correlate_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 
 //      How to do a correlate-style CLASSIFY on some text.
 //
-int crm_expr_correlate_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
+int crm_expr_correlate_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb,
+				 char *txtptr, long txtstart, long txtlen)
 {
 
   //      classify the sparse spectrum of this input window 
@@ -241,13 +254,11 @@ int crm_expr_correlate_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   long i, j, k;
   char ptext[MAX_PATTERN];  //  the regex pattern
   long plen;
-  char ltext[MAX_PATTERN];  //  the variable to classify
-  long llen;
   //  the hash file names
   char htext[MAX_PATTERN+MAX_CLASSIFIERS*MAX_FILE_NAME_LEN];  
   long htext_maxlen = MAX_PATTERN+MAX_CLASSIFIERS*MAX_FILE_NAME_LEN;
   long hlen;
-  //  the match statistics variable inbuf
+  //  the match statistics variable
   char stext[MAX_PATTERN+MAX_CLASSIFIERS*(MAX_FILE_NAME_LEN+100)]; 
   long stext_maxlen = MAX_PATTERN+MAX_CLASSIFIERS*(MAX_FILE_NAME_LEN+100); 
   long slen;
@@ -257,7 +268,6 @@ int crm_expr_correlate_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   char fname[MAX_FILE_NAME_LEN];
   long eflags;
   long cflags;
-  long vhtindex;
 
   struct stat statbuf;      //  for statting the hash file
   //regex_t regcb;
@@ -267,10 +277,14 @@ int crm_expr_correlate_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   double cpcorr[MAX_CLASSIFIERS];  // corpus correction factors
   long long linear_hits[MAX_CLASSIFIERS];  // actual hits per classifier
   long long square_hits[MAX_CLASSIFIERS];  // square of runlenths of match
+  long long cube_hits[MAX_CLASSIFIERS];  // cube of runlength matches
+  long long quad_hits[MAX_CLASSIFIERS];  // quad of runlength matches
   long incr_hits[MAX_CLASSIFIERS];  // 1+2+3... hits per classifier
 
   long long total_linear_hits;  // actual total linear hits for all classifiers
   long long total_square_hits;  // actual total square hits for all classifiers
+  long long total_cube_hits;    // actual total cube hits for all classifiers
+  long long total_quad_hits;    // actual total cube hits for all classifiers
   long long total_features;     // total number of characters in the system
 
   long totalhits [MAX_CLASSIFIERS];
@@ -291,19 +305,15 @@ int crm_expr_correlate_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   long fnstart, fnlen;
   long fn_start_here;
   long textoffset;
-  long textmaxoffset;
   long bestseen;
   long thistotal;
 
   if (internal_trace)
     fprintf (stderr, "executing a CLASSIFY\n");
   
-  //           extract the variable name (if present)
-  //
-  crm_get_pgm_arg (ltext, MAX_PATTERN, apb->b1start, apb->b1len);
-  llen = apb->b1len;
-  llen = crm_nexpandvar (ltext, llen, MAX_PATTERN);
-  
+  //          we use the main line txtptr, txtstart, and txtlen now,
+  //          so we don't need to extract anything from the b1start stuff.
+
   //           extract the hash file names
   crm_get_pgm_arg (htext, htext_maxlen, apb->p1start, apb->p1len);
   hlen = apb->p1len;
@@ -344,20 +354,6 @@ int crm_expr_correlate_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
       eflags = 1;
     };
 
-  //   compile the word regex
-/* unused code
-  if ( internal_trace)
-    fprintf (stderr, "\nWordmatch pattern is %s", ptext);
-  i = crm_regcomp (&regcb, ptext, plen, cflags);
-  if ( i > 0)
-    {
-      crm_regerror ( i, &regcb, tempbuf, data_window_size);
-      nonfatalerror ("Regular Expression Compilation Problem:", tempbuf);
-      goto regcomp_failed;
-    };
-*/
-  
-
   
   //       Now, the loop to open the files.  
   bestseen = 0;
@@ -371,6 +367,8 @@ int crm_expr_correlate_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
       cpcorr[i] = 0.0;   // corpus correction factors
       linear_hits[i] = 0;     // linear hits
       square_hits[i] = 0;     // square of the runlength
+      cube_hits[i] = 0;       // cube of the runlength
+      quad_hits[i] = 0;       // quad of the runlength
       incr_hits[i] = 0;      // 1+2+3... hits hits
       totalhits[i] = 0.0;     // absolute hit counts 
       ptc[i] = 0.5;      // priori probability
@@ -410,8 +408,9 @@ int crm_expr_correlate_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 	    {
 	      if (vbar_seen)
 		{
-		  nonfatalerror ("Only one ' | ' allowed in a CLASSIFY. \n" ,
-				 "We'll ignore it for now.");
+		  nonfatalerror5 ("Only one ' | ' allowed in a CLASSIFY. \n" ,
+				  "We'll ignore it for now.",
+				  CRM_ENGINE_HERE);
 		}
 	      else
 		{
@@ -427,8 +426,8 @@ int crm_expr_correlate_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 	      //             quick check- does the file even exist?
 	      if (k != 0)
 		{
-		  nonfatalerror ("Nonexistent Classify table named: ",
-				 fname);
+		  nonfatalerror5 ("Nonexistent Classify table named: ",
+				 fname, CRM_ENGINE_HERE);
 		}
 	      else
 		{
@@ -436,11 +435,15 @@ int crm_expr_correlate_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 		  //    
 		  hashlens[maxhash] = statbuf.st_size;
 		  hashes[maxhash] = (char *)
-		      crm_mmap_file( 0, fname, hashlens[maxhash] );
-		  if (hashes[maxhash] == NULL )
+		      crm_mmap_file( fname, 
+				     0, hashlens[maxhash],
+				     PROT_READ,
+				     MAP_SHARED,
+				     NULL);
+		  if (hashes[maxhash] == MAP_FAILED )
 			{
-			  nonfatalerror ("Couldn't memory-map the table file",
-					 fname);
+			  nonfatalerror5 ("Couldn't memory-map the table file",
+					  fname, CRM_ENGINE_HERE);
 			}
 		      else
 			{
@@ -454,8 +457,8 @@ int crm_expr_correlate_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 			    //(hashes[maxhash][0].hash != 1 ||
 			    //  hashes[maxhash][0].key  != 0) 
 			    {
-			      fev =fatalerror ("The .css file is the wrong version!  Filename is: ",
-					       &htext[i]);
+			      fev =fatalerror5 ("The .css file is the wrong version!  Filename is: ",
+						&htext[i], CRM_ENGINE_HERE);
 			      return (fev);
 			    };
 
@@ -464,8 +467,10 @@ int crm_expr_correlate_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 			  //
 			  hashname[maxhash] = (char *) malloc (fnlen+10);
 			  if (!hashname[maxhash])
-			    untrappableerror(
-			      "Couldn't malloc hashname[maxhash]\n","We need that part later, so we're stuck.  Sorry.");
+			    untrappableerror5
+			      ("Couldn't malloc hashname[maxhash]\n",
+			       "We need that part later, so we're stuck. "
+			       "Sorry.", CRM_ENGINE_HERE);
 			  strncpy(hashname[maxhash],fname,fnlen);
 			  hashname[maxhash][fnlen]='\000';
 			  maxhash++;
@@ -473,8 +478,9 @@ int crm_expr_correlate_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 		};
 	    };
 	  if (maxhash > MAX_CLASSIFIERS-1)
-	    nonfatalerror ("Too many classifier files.",
-			   "Some may have been disregarded");
+	    nonfatalerror5 ("Too many classifier files.",
+			    "Some may have been disregarded",
+			    CRM_ENGINE_HERE);
 	};
     };
 
@@ -496,7 +502,8 @@ int crm_expr_correlate_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   //	do we have at least 1 valid .css files?
   if (maxhash == 0)
     {
-      fatalerror ("Couldn't open at least 1 .css files for classify().", "");
+      fatalerror5 ("Couldn't open at least 1 .css files for classify().", 
+		   "", CRM_ENGINE_HERE);
     };
   //	do we have at least 1 valid .css file at both sides of '|'?
   //if (!vbar_seen || succhash < 0 || (maxhash < succhash + 2))
@@ -514,31 +521,9 @@ int crm_expr_correlate_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   k = 0;
   thistotal = 0;
 
-  if (llen > 0)
-    {
-      vhtindex = crm_vht_lookup (vht, ltext, llen );
-    }
-  else
-    {
-      vhtindex = crm_vht_lookup (vht, ":_dw:", 5);
-    }
-  if (vht[vhtindex] == NULL)
-    {
-      return (fatalerror (" Attempt to CLASSIFY from a nonexistent variable ",
-			  ltext));
-    };
-  mdw = NULL;
-  if (tdw->filetext == vht[vhtindex]->valtxt)
-    mdw = tdw;
-  if (cdw->filetext == vht[vhtindex]->valtxt)
-    mdw = cdw;
-  if (mdw == NULL)
-    return ( fatalerror (" Bogus text block containing variable ", ltext));  
-  textoffset = vht[vhtindex]->vstart;
-  textmaxoffset = textoffset + vht[vhtindex]->vlen;
-
-  textlen = (vht[vhtindex]->vlen);
-  if (textlen < 1.0) textlen = 1.0; 
+  //     put in the ptr/start/len values we got from the outside caller
+  textoffset = txtstart;
+  textlen = txtlen;
 
   //
   //    We keep track of the hits in these categories
@@ -560,24 +545,30 @@ int crm_expr_correlate_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
       long ik;    // ik is the start index into the known corpus text 
       long ilm;   // ilm is the "local" matches (N in a row)
 
-      //    for each displacement of the test text...
+      //    for each possible displacement of the known  (ik) text...
       for (ik = 0; 
-	   ik < (hashlens[k] - vht[vhtindex]->vlen);
+	   ik < hashlens[k];
 	   ik++)
 	{
+	  long itmax;
 	  ilm = 0;
-	  // for each position in the test text...
+	  itmax = textlen;
+	  if (ik + itmax > hashlens[k])
+	      itmax = hashlens[k] - ik;
+	  // for each position in the test (it) text...
 	  for (it = 0;
-	       it < vht[vhtindex]->vlen;
+	       it < itmax;
 	       it++)
 	    {
 	      //   do the characters in this position match?
-	      if ( hashes[k][ik+it] == mdw->filetext[textoffset+it])
+	      if ( hashes[k][ik+it] == txtptr[textoffset+it])
 		{
 		  // yes they matched
 		  linear_hits[k]++;
 		  ilm++;
 		  square_hits[k] = square_hits[k] + (ilm * ilm);
+		  cube_hits[k] = cube_hits[k] + (ilm * ilm * ilm);
+		  quad_hits[k] = quad_hits[k] + (ilm * ilm * ilm * ilm);
 		}
 	      else
 		{
@@ -586,13 +577,14 @@ int crm_expr_correlate_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 		  ilm = 0;
 		};
 	      if (0)
-		fprintf (stderr, "ik: %ld  it: %ld  c1: %c  c2: %c  ilm: %ld  lin: %lld  sqr: %lld\n",
+		fprintf (stderr, "ik: %ld  it: %ld  chars %c %c lin: %lld  sqr: %lld cube: %lld quad: %lld\n",
 			 ik, it,
 			 hashes[k][ik+it],
-			 mdw->filetext[textoffset+it],
-			 ilm, 
+			 txtptr[textoffset+it],
 			 linear_hits[k],
-			 square_hits[k]);
+			 square_hits[k],
+			 cube_hits[k],
+			 quad_hits[k]);
 	    };
 	};
     };
@@ -617,18 +609,22 @@ int crm_expr_correlate_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   //    Calculate total hits
   total_linear_hits = 0;
   total_square_hits = 0;
+  total_cube_hits = 0;
+  total_quad_hits = 0;
   total_features = 0;
   for (k = 0; k < maxhash; k++)
     {
       total_linear_hits += linear_hits[k];
       total_square_hits += square_hits[k];
+      total_cube_hits += cube_hits[k];
+      total_quad_hits += quad_hits[k];
       total_features += hashlens[k];
     };
 
 
   for (k = 0; k < maxhash; k++)
     {
-      if (hashlens > 0
+      if (hashlens[k] > 0
 	  && total_features > 0 )
 	{
 	  //     Note that we don't normalize the probabilities yet- we do
@@ -639,9 +635,12 @@ int crm_expr_correlate_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 	  //     .00806 is the random coincidence rate for 7-bit characters.
 	  //     
 	  //ptc[k] = ((0.0+square_hits[k] - (.00397 * hashlens[k] )));
-	  ptc[k] = ((0.0+square_hits[k] - (.00806 * hashlens[k] )))
-		    / hashlens[k];
-
+	  //	  ptc[k] = ((0.0+square_hits[k] - (.00806 * hashlens[k] )))
+	  //	    / hashlens[k];
+	  
+	  //	  ptc[k] = (0.0+square_hits[k] ) / hashlens[k];
+	  //	  ptc[k] = (0.0+ quad_hits[k] ) / hashlens[k];
+	  ptc[k] = (0.0+ quad_hits[k] ) / linear_hits[k];
 
 	  if (ptc[k] < 0) ptc[k] = 10*DBL_MIN;
 	}
@@ -778,11 +777,14 @@ int crm_expr_correlate_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 	      };
 	  sprintf (buf, 
 		   "#%ld (%s):"\
-		   " features: %ld, RMS hits: %lld, prob: %3.2e, pR: %6.2f \n", 
+		   " features: %ld, L1: %lld L2: %lld L3: %lld, l4: %lld prob: %3.2e, pR: %6.2f \n", 
 		   k,
 		   hashname[k],
 		   hashlens[k],
+		   linear_hits[k],
 		   square_hits[k],
+		   cube_hits[k],
+		   quad_hits[k],
 		   ptc[k], 
 		   (log10 (ptc[k]) - log10 (remainder) )  );
 	  // strcat (stext, buf);
@@ -794,10 +796,10 @@ int crm_expr_correlate_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
       // whining on stderr
       if (strcmp(&(stext[strlen(stext)-strlen(buf)]), buf) != 0)
         {
-          nonfatalerror( "WARNING: not enough room in the buffer to create "
+          nonfatalerror5( "WARNING: not enough room in the buffer to create "
 			 "the statistics text.  Perhaps you could try bigger "
 			 "values for MAX_CLASSIFIERS or MAX_FILE_NAME_LEN?",
-			 " ");
+			  " ", CRM_ENGINE_HERE);
 	};
       crm_destructive_alter_nvariable (svrbl, svlen, 
 				       stext, strlen (stext));
@@ -808,7 +810,7 @@ int crm_expr_correlate_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   //
   for (i = 0; i < maxhash; i++)
     free (hashname[i]);
-  if (tprob < 0.5000)
+  if (tprob <= 0.5000)
     {
       if (user_trace)
 	fprintf (stderr, "CLASSIFY was a FAIL, skipping forward.\n");
