@@ -1,5 +1,5 @@
-//  crm114_.c  - Controllable Regex Mutilator,  version v1.0
-//  Copyright 2001-2004  William S. Yerazunis, all rights reserved.
+//  crm_exec_engine.c  - Controllable Regex Mutilator,  version v1.0
+//  Copyright 2001-2006  William S. Yerazunis, all rights reserved.
 //  
 //  This software is licensed to the public under the Free Software
 //  Foundation's GNU GPL, version 2.  You may obtain a copy of the
@@ -59,8 +59,9 @@ int crm_invoke ()
   //    Sanity check - don't try to execute a file before compilation
   if (csl->mct == NULL)
     {
-      untrappableerror ( "Can't execute a file without compiling first.\n",
-		   "This means that CRM114 is somehow broken.");
+      untrappableerror5 ( "Can't execute a file without compiling first.\n",
+			  "This means that CRM114 is somehow broken.", 
+			  CRM_ENGINE_HERE);
     };
 
   //    empty out the alius stack (nothing FAILed yet.)
@@ -114,6 +115,14 @@ int crm_invoke ()
       fprintf (stderr, "\n");
     };
 
+
+  //    THIS IS THE ULTIMATE SCREAMING TEST - CHECK VHT EVERY LOOP
+  //     TURN THIS ON ONLY IN EXTREMIS!
+  // do a GC on the whole tdw:
+  //       crm_compress_tdw_section (tdw->filetext, 0, tdw->nchars);
+  
+
+
   //   Invoke the common declensional parser on the statement only if it's
   //   an executable statement.
   //
@@ -122,7 +131,6 @@ int crm_invoke ()
       //
       //   Do the processing that all statements need (well, _almost_ all.)
       //
-    case CRM_LABEL:
     case CRM_NOOP:
     case CRM_BOGUS:
 	break;
@@ -175,7 +183,14 @@ int crm_invoke ()
     {
       i = crm_debugger ();
       if (i == -1)
-	exit ( EXIT_SUCCESS );
+	{
+	  if (engine_exit_base != 0)
+	    {
+	      exit (engine_exit_base + 6);
+	    }
+	  else
+	    exit ( EXIT_SUCCESS );
+	};
       if (i == 1)
 	goto invoke_top;
     };
@@ -231,7 +246,7 @@ int crm_invoke ()
 		csl->mct[csl->cstmt+1]->start - csl->mct[csl->cstmt]->start);
 	bogusstmt [csl->mct[csl->cstmt+1]->start - csl->mct[csl->cstmt]->start]
 	  = '\000';
-        fatalerror (bogusbuffer, bogusstmt);
+        fatalerror5 (bogusbuffer, bogusstmt, CRM_ENGINE_HERE);
 	goto invoke_bailout;
       }
       break;
@@ -240,9 +255,9 @@ int crm_invoke ()
       {
 	int retval; long retlen;
 	char retstr [MAX_PATTERN];
-	crm_get_pgm_arg (retstr, MAX_VARNAME, apb->s1start, apb->s1len);
+	crm_get_pgm_arg (retstr, MAX_PATTERN, apb->s1start, apb->s1len);
 	retlen = apb->s1len;
-	retlen = crm_nexpandvar (retstr, retlen, MAX_VARNAME);
+	retlen = crm_nexpandvar (retstr, retlen, MAX_PATTERN);
 	retval = 0;
 	if (retlen > 0)
 	  sscanf (retstr, "%d", &retval);
@@ -282,7 +297,8 @@ int crm_invoke ()
 	//     Evaluate the "subject", and return that.  
 	if (apb->s1len > 0)
 	  {
-	    unsigned long idx, noffset;
+	    long idx;
+	    unsigned long noffset;
 	    crm_get_pgm_arg (outbuf, MAX_VARNAME, apb->s1start, apb->s1len);
 	    retlen = apb->s1len;
 	    retlen = crm_nexpandvar (outbuf, retlen, data_window_size);
@@ -297,6 +313,7 @@ int crm_invoke ()
 	      }
 	    else
 	      {
+		long i;
 		//fprintf (stderr, "idx: %lu  vht[idx]: %lu", idx, vht[idx] );
 		noffset = vht[idx]->nstart;
 		//fprintf (stderr, " idx: %lu noffset: %lu \n/", idx, noffset);
@@ -305,8 +322,12 @@ int crm_invoke ()
 		namelen = vht[idx]->nlen;
 		
 		if (user_trace)
-		  fprintf (stderr, " setting return value.of >%*s<\n",
-			   (int)namelen, namestart );
+		  {
+		    fprintf (stderr, " setting return value of >");
+		    for (i = 0; i < namelen; i++ )
+		      fprintf (stderr, "%c", namestart[i]);
+		    fprintf (stderr, "<\n");
+		  }
 		//     stuff the return value into that variable.
 		//
 		crm_destructive_alter_nvariable (namestart, namelen,
@@ -331,23 +352,24 @@ int crm_invoke ()
       
     case CRM_GOTO:
       {
-	char target[MAX_VARNAME];
+	static char target[MAX_VARNAME];
 	long tarlen;
 	//  look up the variable name in the vht.  If it's not there, or 
 	//  not in our file, call a fatal error.
 
 	crm_get_pgm_arg ( target, MAX_VARNAME, apb->s1start, apb->s1len);
-	if (apb->s1len < 1)
-	  nonfatalerror 
+	if (apb->s1len < 2)
+	  nonfatalerror5 
 	    ("This program has a GOTO without a place to 'go' to.",
-	     " By any chance, did you leave off the '/' delimiters? ");
+	     " By any chance, did you leave off the '/' delimiters? ",
+	     CRM_ENGINE_HERE);
 	tarlen = apb->s1len;
 	if (internal_trace) 
 	  fprintf (stderr, "\n    untranslated label %s , ", 
 		   target);
 
 	//   do indirection if needed.
-	tarlen = crm_nexpandvar (target, tarlen, MAX_VARNAME);
+	tarlen = crm_qexpandvar (target, tarlen, MAX_VARNAME, NULL);
 	if (internal_trace)
 	  fprintf (stderr, " translates to %s .", target);
 
@@ -358,16 +380,32 @@ int crm_invoke ()
 	    if (user_trace)
 	      fprintf (stderr, "GOTO from line %ld to line %ld\n", 
 		       csl->cstmt,  k);
-	    csl->cstmt = k;
+	    csl->cstmt = k;  // this gets autoincremented
 	    //  and going here didn't fail... 
 	    csl->aliusstk [ csl->mct[csl->cstmt]->nest_level ] = 1;
 	  }
 	else
 	  {
-	    //  this is recoverable if we have a trap... so we continue
-	    //   execution right to the BREAK.
-	    fatalerror (" Can't GOTO the nonexistent label: ", target);
-	    goto invoke_bailout;
+	    int conv_count;
+	    conv_count = sscanf (target, "%ld", &k);
+	    if (conv_count == 1)
+	      {
+		if (user_trace)
+		  fprintf (stderr, "GOTO from line %ld to line %ld\n",
+			   csl->cstmt, k);
+		csl->cstmt = k -1 ; // this gets autoincremented, so we must --
+		//  and going here didn't fail...
+		csl->aliusstk [ csl->mct[csl->cstmt]->nest_level ] = 1;
+	      }
+	    else
+	      {
+		//  this is recoverable if we have a trap... so we continue
+		//   execution right to the BREAK.
+		fatalerror5 (" Can't GOTO the nonexistent label/line: ", 
+			     target,
+			     CRM_ENGINE_HERE);
+		goto invoke_bailout;
+	      };
 	  };
       }
       break;
@@ -465,8 +503,9 @@ int crm_invoke ()
 	fresult = crm_trigger_fault (reason);
 	if (fresult != 0)
 	  {
-	    fatalerror("Your program has no TRAP for the user defined fault:",
-			reason);
+	    fatalerror5 
+	      ("Your program has no TRAP for the user defined fault:",
+	       reason, CRM_ENGINE_HERE);
 	    goto invoke_bailout;
 	  };
       }
@@ -490,8 +529,9 @@ int crm_invoke ()
 	if (varidx == 0
 	    || vht[varidx] == NULL)
 	  {
-	    fatalerror ("This is very strange... there is no data window!",
-			"As such, death is our only option.");
+	    fatalerror5 ("This is very strange... there is no data window!",
+			 "As such, death is our only option.",
+			 CRM_ENGINE_HERE );
 	    goto invoke_bailout;
 	  }
 	else
@@ -539,7 +579,7 @@ int crm_invoke ()
 	crm_expr_eval (csl, apb);
       }
       break;
-      
+
     case CRM_HASH:
       {
 	//      here's where we surgiclly alter a variable to a hash.
@@ -601,6 +641,11 @@ int crm_invoke ()
       };
       break;
 
+    case CRM_TRANSLATE:
+      {
+	crm_expr_translate (csl, apb);
+      };
+      break;
 
     case CRM_LEARN:
       {
@@ -614,7 +659,9 @@ int crm_invoke ()
       break;
 
     case CRM_ISOLATE:
-      crm_expr_isolate (csl, apb);
+      //    nonzero return means "bad things happened"...
+      if ( crm_expr_isolate (csl, apb) )
+	goto invoke_bailout;
       break;
       
     case CRM_INPUT:
@@ -625,7 +672,8 @@ int crm_invoke ()
 
     case CRM_SYSCALL:
       {
-	crm_expr_syscall (csl, apb);
+	if (crm_expr_syscall (csl, apb))
+	  goto invoke_bailout;
       }
       break;
 
@@ -661,7 +709,8 @@ int crm_invoke ()
 	  {
 	    //    aw, crud.  No such label known.  But it _is_ continuable
 	    //    if there is a trap for it.
-	    fatalerror (" Can't CALL the nonexistent label: ", target);
+	    fatalerror5 (" Can't CALL the nonexistent label: ", target,
+			 CRM_ENGINE_HERE);
 	    goto invoke_bailout;
 	  }
 	newcsl = (CSL_CELL *) malloc (sizeof (CSL_CELL));
@@ -676,8 +725,9 @@ int crm_invoke ()
 	newcsl->preload_window = csl->preload_window;
 	newcsl->caller = csl;
 	newcsl->calldepth = csl->calldepth + 1;
-	//     put in the target statement number
+	//     put in the target statement number - this is a label!
 	newcsl->cstmt = k;
+	newcsl->return_vht_cell = -1;
 	//     whack the alius stack so we are not in a "fail" situation
 	newcsl->aliusstk [csl->mct[csl->cstmt]->nest_level + 1 ] = 0;
 	
@@ -713,25 +763,37 @@ int crm_invoke ()
 
 	  //   Stuff the new csl with the return-value-locations'
 	  //   vht index - if it's -1, then we don't have a return
-	  //   value location.
+	  //   value location.  We do this now rather than on return
+	  //   so that we already have the vht entry rather than a 
+	  //   name, and so the returnname isn't dependent on the 
+	  //   function being executed (is this a bug?  What if the
+	  //   returnname is :+:something: ?)
 	  //
 	  newcsl->return_vht_cell = -1;
 	  if (apb->p1len > 0)
 	    {
 	      unsigned long ret_idx;
-	      unsigned long retnamelen;
+	      long retname_start, retnamelen;
 	      crm_get_pgm_arg (outbuf, data_window_size,
 			       apb->p1start, apb->p1len);
 	      retnamelen = apb->p1len;
 	      retnamelen = crm_nexpandvar (outbuf, 
 					   retnamelen, data_window_size);
-	      ret_idx = crm_vht_lookup (vht, outbuf, retnamelen);
+	      crm_nextword (outbuf, retnamelen,0, &retname_start, &retnamelen);
+	      ret_idx = crm_vht_lookup (vht, 
+					&outbuf[retname_start], 
+					retnamelen);
 	      if (vht[ret_idx] == NULL)
 		{
-		  nonfatalerror ("Your call statement wants to return a value "
-				 "to a nonexistent variable; I'll created an "
-				 "isolated one.  Hope that's OK. Varname was",
-				 outbuf);
+		  // nonfatalerror 
+		  // ("Your call statement wants to return a value "
+		  // "to a nonexistent variable; I'll created an "
+		  //"isolated one.  Hope that's OK. Varname was",
+		  //		 outbuf);
+		  if (user_trace)
+		    fprintf (stderr, 
+			     "No such return value var, creating var %s\n",
+			     outbuf);
 		  crm_set_temp_var( outbuf, "");
 		};
 	      ret_idx = crm_vht_lookup (vht, outbuf, retnamelen);
@@ -898,8 +960,10 @@ int crm_invoke ()
 		    char varname[MAX_VARNAME];
 		    strncpy (varname, &temp_vars[vstart], vlen);
 		    varname[vlen] = '\000';
-		    nonfatalerror ( "can't intersection a nonexistent variable.",
-				    varname);
+		    nonfatalerror5 
+		      ( "can't intersection a nonexistent variable.",
+			varname,
+			CRM_ENGINE_HERE);
 		    goto invoke_bailout;
 		  }
 		else
@@ -911,8 +975,8 @@ int crm_invoke ()
 			char varname[MAX_VARNAME];
 			strncpy (varname, &temp_vars[vstart], vlen);
 			varname[vlen] = '\000';
-			nonfatalerror ( "can't intersect isolated variable.",
-					varname);
+			nonfatalerror5 ( "can't intersect isolated variable.",
+					 varname, CRM_ENGINE_HERE);
 			goto invoke_bailout;
 
 		      }
@@ -1008,8 +1072,8 @@ int crm_invoke ()
 		    char varname[MAX_VARNAME];
 		    strncpy (varname, &temp_vars[vstart], vlen);
 		    varname[vlen] = '\000';
-		    nonfatalerror ( "can't intersect a nonexistent variable.",
-				    varname);
+		    nonfatalerror5 ( "can't intersect a nonexistent variable.",
+				     varname, CRM_ENGINE_HERE);
 		    goto invoke_bailout;
 
 		  }
@@ -1022,8 +1086,8 @@ int crm_invoke ()
 			char varname[MAX_VARNAME];
 			strncpy (varname, &temp_vars[vstart], vlen);
 			varname[vlen] = '\000';
-			nonfatalerror ( "can't intersect isolated variable.",
-					varname);
+			nonfatalerror5 ( "can't intersect isolated variable.",
+					 varname, CRM_ENGINE_HERE);
 			goto invoke_bailout;
 		      }
 		    else
@@ -1052,6 +1116,27 @@ int crm_invoke ()
 	
       }
       break;
+      
+     
+    case CRM_DEBUG:
+      {               // turn on the debugger - NOW!
+	if (user_trace)
+	  fprintf (stderr, "executing a DEBUG statement - drop to debug\n");
+	debug_countdown = 0;
+      }
+      break;
+
+    case CRM_CLUMP:
+      {
+	crm_expr_clump (csl, apb);
+      };
+      break;
+
+    case CRM_PMULC:
+      {
+	crm_expr_pmulc (csl, apb);
+      };
+      break;
 
     case CRM_UNIMPLEMENTED:
       {
@@ -1065,7 +1150,7 @@ int crm_invoke ()
 		csl->mct[csl->cstmt+1]->start - csl->mct[csl->cstmt]->start);
 	bogusstmt [csl->mct[csl->cstmt+1]->start - csl->mct[csl->cstmt]->start]
 	  = '\000';
-        fatalerror (bogusbuffer, bogusstmt);
+        fatalerror5 (bogusbuffer, bogusstmt, CRM_ENGINE_HERE);
 	goto invoke_bailout;
 
       };
@@ -1083,7 +1168,7 @@ int crm_invoke ()
 		csl->mct[csl->cstmt+1]->start - csl->mct[csl->cstmt]->start);
 	bogusstmt [csl->mct[csl->cstmt+1]->start - csl->mct[csl->cstmt]->start]
 	  = '\000';
-        fatalerror (bogusbuffer, bogusstmt);
+        fatalerror5 (bogusbuffer, bogusstmt, CRM_ENGINE_HERE);
 	goto invoke_bailout;
       };
     }
