@@ -1,5 +1,5 @@
-//  crm114_.c  - Controllable Regex Mutilator,  version v1.0
-//  Copyright 2001-2004  William S. Yerazunis, all rights reserved.
+//  crm_expandvar.c  - Controllable Regex Mutilator,  version v1.0
+//  Copyright 2001-2006  William S. Yerazunis, all rights reserved.
 //  
 //  This software is licensed to the public under the Free Software
 //  Foundation's GNU GPL, version 2.  You may obtain a copy of the
@@ -119,10 +119,13 @@ long crm_zexpandvar (char *buf,
   long is, id;
   long vht_index;
   long q;
+
+  //  the maximum length allocated so far for these random buffers...
+  static long current_maxlen = 0;
   //  a temporary work buffer...
-  char *tbuf;
+  static char *tbuf = NULL;
   //  and another for variable names...
-  char *vname; 
+  static char *vname = NULL; 
   
   char *cp;
   long vlen;
@@ -144,9 +147,9 @@ long crm_zexpandvar (char *buf,
   //  GROT GROT GROT must fix this for 8-bit safe error messages
   if (inlen > maxlen)
     {
-      q = fatalerror (
-		      "You have blown the gaskets while building a string.  Orig string was: ",
-		      buf);
+      q = fatalerror5 
+	( "You have blown the gaskets while building a string.  Orig string was: ",
+	  buf, CRM_ENGINE_HERE);
       if (q == 0 )
 	return (inlen);
       goto bailout;
@@ -331,24 +334,54 @@ long crm_zexpandvar (char *buf,
   if (cp == NULL)
     {
       if (internal_trace)
-	fprintf (stderr, "No further expansions possible");
+	fprintf (stderr, "No further expansions possible\n");
       return (inlen);
     };
   
-  //    OK, we might have a :*: substitution operator, so we actually have
-  //    to do some work.
-  //    allocate some memory for tbuf and vname;
-  tbuf = (char *) malloc (maxlen+1);
-  vname = (char *) malloc (maxlen+1);
+
+  //    allocate some memory for tbuf and vname; (this funky allocation
+  //    is a workaround for malloc memory fragmentation that caused
+  //    out-of-memory problems in some kernels.  Eventually we'll have
+  //    a much grander system for all mallocs, but not yet.)
+
+  //   if the currently allocated buffers are too small, drop them
+  //   (and force a reallocation), else we will reuse them.
+  if (current_maxlen < maxlen + 1)     // do we need to drop and reallocate? 
+    {
+      if (tbuf != NULL) 
+	{
+	  free (tbuf);
+	  tbuf = NULL;
+	};
+      if (vname != NULL) 
+	{
+	  free (vname);
+	  vname = NULL;
+	};
+      current_maxlen = maxlen + 2;
+    }
   
+  if (tbuf == NULL )
+    {
+      tbuf = (char *) malloc (current_maxlen);
+    };
+  
+  if (vname == NULL)
+    vname = (char *) malloc (current_maxlen);
+ 
+
   if (tbuf == NULL || vname == NULL)
     {
-      q = fatalerror ("Couldn't allocate memory for Q-variable expansion!",
-		      "Try making the window set smaller with the -w option");
+      q = fatalerror5 ("Couldn't allocate memory for Q-variable expansion!",
+		      "Try making the window set smaller with the -w option",
+		      CRM_ENGINE_HERE);
       if (q == 0)
 	return (inlen);
     };
   
+  //    OK, we might have a :*: substitution operator, so we actually have
+  //    to do some work.
+
   //
   //    Now, do we have a :*: (singlevar) possible?
 
@@ -437,9 +470,20 @@ long crm_zexpandvar (char *buf,
 				  (char *) &vht[vht_index]->nametxt[vht[vht_index]->nstart],
 				  ":_iso:", 6) == 0)
 			{
-			  //   if this was :_:, update iso's length
+			  //   if this was :_iso:, update iso's length
 			  vht[vht_index]->vlen = tdw->nchars;
 			};
+		      if (strncmp(
+				  (char *) &vht[vht_index]->nametxt[vht[vht_index]->nstart],
+				  ":_cs:", 5) == 0)
+			{
+			  //   if this was :_cs:, update the current line num
+			  char lcstring [32];
+			  long lclen;
+			  lcstring[0] = '\0';
+			  lclen = sprintf (lcstring, "%ld", csl->cstmt);
+			  crm_set_temp_nvar (":_cs:", lcstring, lclen); 
+			};
 		    };
 		  
 		  switch (opchar)
@@ -467,9 +511,13 @@ long crm_zexpandvar (char *buf,
 	}
 
       //    and put our results back into buf       
-      strncpy (buf, tbuf, id);
+      memcpy (buf, tbuf, id);
       buf[id] = '\000';
       inlen = id - 1;
+
+      if (internal_trace)
+	fprintf (stderr, " :*: var-expansion yields: =%s= len %ld \n", buf, inlen);
+
     }  
 
   //     END OF :*: EXPANSIONS
@@ -484,7 +532,7 @@ long crm_zexpandvar (char *buf,
       is = 0;    //   is is the input position index
       id = 0;    //   id is the destination position index
       if (internal_trace)
-	fprintf (stderr, "Doing singlevar eval expansion\n");
+	fprintf (stderr, "Doing singlevar redirect expansion\n");
 
       //
       //   First time through the loop, for :+: (variable expansion)
@@ -603,9 +651,11 @@ long crm_zexpandvar (char *buf,
 	    }
 	}
       //    and put our results back into buf       
-      strncpy (buf, tbuf, id);
+      memcpy (buf, tbuf, id);
       buf[id] = '\000';
       inlen = id - 1;
+      if (internal_trace)
+	fprintf (stderr, "indirection :+: expansion yields: =%s= len %ld \n", buf, inlen);
     }  
   
   //     END OF :+: EXPANSIONS
@@ -615,7 +665,7 @@ long crm_zexpandvar (char *buf,
     {
       //
       //
-      //   Second time through the loop - expand :#: (string lengths)
+      //   Expand :#: (string lengths)
       //
       if (internal_trace)
 	fprintf (stderr, "Doing stringglength expansion\n");
@@ -729,10 +779,13 @@ long crm_zexpandvar (char *buf,
 	    }
 	}
       //    and put our results back into buf       
-      strncpy (buf, tbuf, id);
+      memcpy (buf, tbuf, id);
       buf[id] = '\000';
       //    and because id always gets an extra increment...
       inlen = id - 1;
+      if (internal_trace)
+	fprintf (stderr, " strlen :#: expansion yields: =%s= len %ld \n", buf, inlen);
+
     }
   //       END OF :#: STRING LENGTH EXPANSIONS
 
@@ -784,8 +837,9 @@ long crm_zexpandvar (char *buf,
 		}
 	      else
 		{
-		  nonfatalerror ("This math eval didn't end with a ':' which is",
-				 " often an error...  check it sometime? ");
+		  nonfatalerror5 ("This math eval didn't end with a ':' which is",
+				  " often an error...  check it sometime? ",
+				  CRM_ENGINE_HERE);
 		};
 	      vname [vlen] = '\000';
 	      
@@ -805,7 +859,7 @@ long crm_zexpandvar (char *buf,
 		      {
 			char mathtext[MAX_VARNAME];
 			int m, mm;
-			strncpy (mathtext, &vname[1], vlen-2);
+			memcpy (mathtext, &vname[1], vlen-2);
 			mathtext[vlen-2] = '\000';
 			if (internal_trace)
 			  fprintf (stderr, "In-Mathtext is -'%s'-\n", mathtext); 
@@ -814,8 +868,9 @@ long crm_zexpandvar (char *buf,
 			  fprintf (stderr, "Out-Mathtext is -'%s'-\n", mathtext);
 			if (retstat && *retstat < 0)
 			  {
-			    q = fatalerror ("Problem during math evaluation of ",
-					    mathtext);
+			    q = fatalerror5 
+			      ("Problem during math evaluation of ",
+			       mathtext, CRM_ENGINE_HERE);
 			    if (q == 0)
 			      return (inlen);
 			    goto bailout;
@@ -860,8 +915,9 @@ long crm_zexpandvar (char *buf,
 			m = strmath (mathtext, vlen-2, MAX_VARNAME, retstat );
 			if (retstat && *retstat < 0)
 			  {
-			    q = fatalerror ("Problem during math evaluation of ",
-					    mathtext);
+			    q = fatalerror5 
+			      ("Problem during math evaluation of ",
+			       mathtext, CRM_ENGINE_HERE);
 			    if (q == 0)
 			      return (inlen);
 			    goto bailout;
@@ -886,7 +942,7 @@ long crm_zexpandvar (char *buf,
 	    }
 	}
       //    and put our results back into buf       
-      strncpy (buf, tbuf, id);
+      memcpy (buf, tbuf, id);
       buf [id] = '\000';
       inlen = id - 1;
 
@@ -900,8 +956,10 @@ long crm_zexpandvar (char *buf,
   //    That's all, folks!  Clean up the temporary buffer.  We null-terminate
   //    it in case we need to do stupid non-8-bit-clean IO on it.
   tbuf[inlen+1] = '\000';
-  free (tbuf);
-  free (vname);
+  
+  //     We reuse tbuf and vname from now on.
+  // free (tbuf);
+  //free (vname);
   if (internal_trace)
     {
       fprintf (stderr, " Returned length from qexpandvar is %ld\n", inlen);
@@ -913,3 +971,322 @@ long crm_zexpandvar (char *buf,
   return (inlen);
 }
 
+////////////////////////////////////////////////////////////////////
+//
+//     crm_restrictvar - hand this routine a []-string, and it hands
+//     you back the VHT index, the applicable char* for the start, and
+//     the length.  Or maybe an error.  
+//
+//     Error codes: 0 == none, 1 = nonfatal, 2 = fatal
+//
+//     Algorithm: first "nextword" thing is always the var.  Grab it,
+//     and get the pertinent info from the VHT.  Successive nextwords
+//     get either a /regex/ or a numeric (or possibly a numeric pair).
+//     On each one, do successive regexing/indexranging.  When no more
+//     nextwords, you're done.
+//     
+
+long crm_restrictvar ( char *boxstring,
+                       long boxstrlen,
+		       long *vht_idx,
+                       char **outblock,
+                       long *outoffset,
+		       long *outlen,
+                       char *errstr)
+{
+  char datastring[MAX_PATTERN + 1];
+  long datastringlen;
+
+  char varname[MAX_PATTERN + 1];
+  long varnamelen;
+  long vmidx;
+
+  regex_t preg;
+  regmatch_t matches [MAX_SUBREGEX];
+
+  char scanbuf[MAX_PATTERN + 1];
+  long scanbuflen;
+  
+  long nw_start, nw_len;
+
+  char *mdw;               //  the data window that this var is stored in.
+  char *start_ptr;
+  long actual_offset;
+  long actual_len;
+  long in_subscript;
+
+  long i, j;
+
+  nw_start = 0;
+  nw_len = 0;
+
+  if (user_trace)
+    fprintf (stderr, "Performing variable restriction.\n");
+  
+  //     Expand the string we were handed.
+  memcpy (datastring, boxstring, boxstrlen);
+  datastring[boxstrlen] = '\0';
+
+  if (user_trace)
+    fprintf (stderr, "Variable before expansion '%s' len %ld\n",
+	     datastring, boxstrlen);
+  datastringlen = crm_qexpandvar(datastring, boxstrlen, MAX_PATTERN, NULL); 
+  if (user_trace)
+    fprintf (stderr, "Variable after expansion: '%s' len %ld\n", 
+	     datastring, datastringlen);
+ 
+  //     Get the variable name.
+  crm_nextword (datastring, datastringlen, nw_start, &nw_start, &nw_len);
+  if (internal_trace)
+    fprintf (stderr, "box-parsing varname got start: %ld, len: %ld .\n",
+	     nw_start, nw_len);
+
+  if (nw_len > 0)
+    {
+      memcpy (varname, &datastring[nw_start], nw_len);
+      varname[nw_len] = '\0';
+      varnamelen = nw_len;
+    }
+  else
+    {
+      //    if no variable, use :_dw:
+      memcpy (varname, ":_dw:", 6);
+      varnamelen = 5;
+    };
+  if (user_trace)
+    fprintf (stderr, "Using variable '%s' for source.\n", varname);
+
+  //      Got the varname.  Do a lookup.
+  vmidx = crm_vht_lookup (vht, varname, varnamelen);
+  if (internal_trace)
+    fprintf (stderr, "vmidx = %ld, vht[vmidx] = %lx\n", 
+	     (long) vmidx, (long) vht[vmidx]);
+  //       Is it a real variable?
+  if ( ((void *) vht[vmidx]) == NULL )  
+    {
+      strcpy ( errstr, 
+	       "This program wants to use a nonexistent variable named: '");
+      strncat ( errstr, varname, MAX_PATTERN - 128);
+      strcat (errstr, "'");
+      return (-2);
+    };
+  
+  //     Get the data window - cdw, or tdw.
+  mdw = cdw->filetext;    //  assume cdw unless otherwise proven...
+  if (vht[vmidx]->valtxt == tdw->filetext)
+    mdw=tdw->filetext;
+  //  sanity check - must be tdw or cdw for searching!
+  if (vht[vmidx]->valtxt != tdw->filetext
+      && vht[vmidx]->valtxt != cdw->filetext)
+    {
+      errstr[0] = '\0';
+      strcat (errstr, "Bogus text block (neither cdw nor tdw) on var ");
+      strcat (errstr, varname);
+      strcat (errstr, "\n");
+      return (-2);
+    };
+
+  if (user_trace)
+    fprintf (stderr, "Found that variable\n");
+  actual_offset = vht[vmidx]->vstart;
+  actual_len = vht[vmidx]->vlen;
+
+  //     Now, we can go through the remaining terms in the var restriction 
+  //     and chop the maximal region down (if desired)
+  
+  in_subscript = 0;
+  while ( nw_start <= datastringlen )
+    {
+      if (user_trace) 
+	fprintf (stderr, 
+		 "Checking restriction at start %ld len %ld (subscr=%ld)\n",
+		 nw_start+nw_len,
+		 (datastringlen - (nw_start + nw_len)),
+		 in_subscript);
+
+      //      get the next word
+      crm_nextword (datastring, datastringlen, nw_start+nw_len, 
+		    &nw_start, &nw_len);
+
+      if (internal_trace)
+	fprintf (stderr, "box-parsing left returned start: %ld, len: %ld .\n",
+		 nw_start, nw_len);
+      
+      //     Are we done?
+      if (nw_len <= 0) 
+	{
+	  if (user_trace)
+	    fprintf (stderr, "Nothing more to do in the var-restrict.\n");
+	  break;
+	}
+
+      //      we need to shred the word (put a NULL at the end so we can 
+      //      use sscanf on it )
+      memcpy (scanbuf, &datastring[nw_start], nw_len);
+      scanbuflen = nw_len;
+      scanbuf[scanbuflen] = '\0';
+            
+      if (internal_trace)
+	fprintf (stderr, "  var restrict clause was '%s' len %ld \n", 
+		 scanbuf, scanbuflen);
+
+      //      Is it int-able?
+      i = sscanf ( scanbuf, "%ld", &j);
+      if (i > 0) 
+	{
+	  //   Check for a negative value of j; negative j would allow
+	  //   out-of-bounds accessing.
+	  if (j < 0)
+	    {  
+	      j = 0;
+	      fprintf (stderr, "Var-restriction has negative start or length."
+		     "  Sorry, but negative start/lengths are not "
+		       "allowed, as it's a possible security exploit.");
+	    };
+	  //      Do the offset/length alternation thing.
+	  if (in_subscript == 0)
+	    {
+	      if (actual_len <= j) 
+		{
+		  if (user_trace) fprintf (stderr, "Clipping start to %ld",
+					   actual_len);
+		  j = actual_len;
+		};
+	      if (user_trace)
+		fprintf (stderr, "Variable starting offset: %ld\n", j);
+	      actual_offset = actual_offset + j;
+	      actual_len = actual_len - j;
+	      in_subscript = 1;
+	    }
+	  else
+	    {
+	      if (actual_len < j) 
+		{
+		  if (user_trace)
+		    fprintf (stderr, "Clipping length to %ld\n", 
+			     actual_len);
+		  j = actual_len;
+		};
+	      if (user_trace)
+		fprintf (stderr, "Variable starting offset: %ld\n", j);
+	      actual_len = j;
+	      in_subscript = 0;
+	    }
+	}
+      else
+	//      it's not an int; see if it's a /regex/
+	{
+	  long regex_start;
+	  in_subscript = 0;   //  no longer in subscript-length mode.
+	  if (datastring[nw_start] == '/')
+	    //      yes, it's a start of regex.  copy it into the scan buf,
+	    //      while looking for the closing '/', and keeping 
+	    //      any \/ as / (everything eles is verbatim).
+	    {
+	      regex_start = nw_start + 1;  // regex starts +1 past start of str
+	      nw_len = 0;                     // nw_len is next open char idx.
+	      while ( (regex_start < datastringlen
+		       && datastring [regex_start] != '/' )
+		      || ( regex_start < datastringlen
+			   && datastring [regex_start] == '/'
+			   && datastring [regex_start-1] == '\\'))
+		{
+		  //   overwrite escaped slashes?
+		  if (datastring[regex_start] == '/')
+		    nw_len--;       
+		  scanbuf[nw_len] = datastring[regex_start];
+		  nw_len++;
+		  regex_start++;
+		}
+	      scanbuf[nw_len] = '\0';
+	      if (user_trace)
+		fprintf (stderr, "Var restriction with regex '%s' len %ld\n", 
+			 scanbuf, nw_len);
+
+	      //
+	      //    Compile up that regex
+	      j = crm_regcomp (&preg, scanbuf, nw_len, REG_EXTENDED);
+	      if (j > 0) 
+		{
+		  long curstmt;
+		  curstmt = csl->cstmt;
+		  crm_regerror ( i, &preg, tempbuf, data_window_size);
+		  strcpy (errstr, 
+			  "Regular Expression Compilation Problem on:");
+		  strncat (errstr, tempbuf, MAX_PATTERN - 128);
+		  return (-2);
+		};
+	      
+	      if (internal_trace)
+		fprintf (stderr, " Running regexec, start at %ld\n",
+			 actual_offset);
+	      //    Time to run the match
+	      start_ptr= &(mdw[actual_offset]);
+	      j = crm_regexec ( &preg, start_ptr, actual_len,
+				MAX_SUBREGEX, matches, 0, NULL);
+	      crm_regfree (&preg);
+	      if (j == 0)
+		{
+		  //    Yes, the regex matched.  Find the innermost 
+		  //    participating submatch, and use that.
+		  long i;
+		  i = 0;
+		  while (matches[i].rm_so >= 0)
+		    i++;
+		  i--;
+		  //     Now use the stuff in matches[i] as 
+		  //    data to seet the new limits to our var
+		  actual_offset = actual_offset + matches[i].rm_so;
+		  actual_len = matches[i].rm_eo - matches[i].rm_so;
+		  if (user_trace)
+		    fprintf (stderr, " Var restrict regex matched, "
+			     "new start offset %ld, new length %ld\n",
+			     (long) matches[i].rm_so, (long) matches[i].rm_eo);
+		}
+	      else
+		{
+		  //    The regex didn't match.  We're done.  Length
+		  //    is now zero.
+		  actual_len = 0;
+		  if (user_trace)
+		    fprintf (stderr, "Var restrict regex didn't match, "
+			     "string is now zero length.\n");
+		  goto all_done;
+		};
+	    }
+	}
+    }
+ all_done:
+  /////////////////////////////
+  //     All calculations done.  Push actual start and actual length
+  //     back onto the output vars
+
+  if ( outblock )
+    *outblock = vht[vmidx]->valtxt;
+  if ( outoffset)
+    *outoffset = actual_offset;
+  if ( outlen )
+    *outlen = actual_len;
+  if (vht_idx)
+    *vht_idx = vmidx;
+  //
+  if (internal_trace)
+    {
+      fprintf (stderr, "Final non-nulls: ");
+      if (vht_idx)
+	fprintf (stderr, " VHTidx %ld", (unsigned long) *vht_idx);
+      if (outblock)
+	fprintf (stderr, " blockaddr %ld", (unsigned long) *outblock);
+      if (outoffset)
+	fprintf (stderr, " startoffset %ld", (unsigned long) *outoffset);
+      if (outlen)
+	fprintf (stderr, " length %ld", (unsigned long) *outlen);
+      fprintf (stderr, "\n");
+    };
+  return ( 0 );
+}
+
+    
+    
+
+
