@@ -1,5 +1,5 @@
 //  crm_expr_match.c  - Controllable Regex Mutilator,  version v1.0
-//  Copyright 2001-2004  William S. Yerazunis, all rights reserved.
+//  Copyright 2001-2006  William S. Yerazunis, all rights reserved.
 //  
 //  This software is licensed to the public under the Free Software
 //  Foundation's GNU GPL, version 2.  You may obtain a copy of the
@@ -43,8 +43,8 @@ int crm_expr_match (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   long mc;
   char pch[MAX_PATTERN];
   long pchlen;
-  char auxstr[MAX_PATTERN];
-  long auxstr_len;
+  char errstr[MAX_PATTERN+128];
+  char *mdwptr;
   regex_t preg;
   int casep, nomultilinep, absentp, fromp, extended_regex_p, literal_pattern_p;
   int cflags, eflags;
@@ -54,13 +54,10 @@ int crm_expr_match (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   char bindable_vars[MAX_PATTERN]; 
   long bindable_vars_len;
   char box_text[MAX_PATTERN];
-  char box_temp[MAX_PATTERN];
-  long box_start;
-  long box_length;
-  char svname[MAX_PATTERN];
-  long svnamelen;
   regmatch_t matches [MAX_SUBREGEX];
   long nmatches;
+  long source_start;
+  long source_len;
   long vtextoffset, vtextend, vtextstartlimit;
   long vpmstart, vpmend;
   long vmidx;
@@ -81,7 +78,7 @@ int crm_expr_match (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 
   casep = 0;
   nomultilinep = 0;
-  absentp = 1;
+  absentp = 0;
   fromp = 0;
   extended_regex_p = 1;
   literal_pattern_p = 0;
@@ -113,7 +110,7 @@ int crm_expr_match (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
     {
       if (user_trace)
 	fprintf (stderr, "  absent flag turned on...\n");
-      absentp = 0;
+      absentp = 1;
     };
 
   if (apb->sflags & CRM_LITERAL)
@@ -192,54 +189,49 @@ int crm_expr_match (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   bindable_vars_len = crm_nexpandvar (bindable_vars, apb->p1len, MAX_PATTERN);
   
   
-  //     here's where we look for a sub-match - specifically,
-  //     to match inside a previously bound variable.
+  //     here's where we look for a [] var-restriction
   //
   //     Experimentally, we're adding [ :foo: 123 456 ] to 
   //     allow an externally specified start and length.
   crm_get_pgm_arg (box_text, MAX_PATTERN, apb->b1start, apb->b1len);
-  i = 0;
-  j = 0;
-  crm_nextword (box_text, apb->b1len, 0, &i, &j);
-  svname[0] ='\0';
-  memmove (svname, &box_text[i], j);
-  svnamelen = crm_nexpandvar (svname, j, MAX_PATTERN);
-  
-  //      now, do we have a start?
-  box_start = 0;
-  crm_nextword (box_text, apb->b1len, i+j, &i, &j);
-  memmove (box_temp, &box_text[i], j);
-  box_temp[j] = '\0';
-  crm_qexpandvar (box_temp, j, MAX_PATTERN, NULL);
-  sscanf (box_temp, "%ld", &box_start);
-  if (user_trace)
-    fprintf (stderr, "Setting box-set starting offset to %ld\n", box_start);
 
-  //      now, do we have a length?
-  box_length = data_window_size;
-  crm_nextword (box_text, apb->b1len, i+j, &i, &j);
-  memmove (box_temp, &box_text[i], j);
-  box_temp[j] = '\0';
-  crm_qexpandvar (box_temp, j, MAX_PATTERN, NULL);
-  sscanf (box_temp, "%ld", &box_length);
-  if (user_trace)
-    fprintf (stderr, "Setting box-set max length to %ld\n", box_length);
+  //  Use crm_restrictvar to get start & length to look at.
+  i = crm_restrictvar(box_text, apb->b1len, 
+		      &vmidx,
+		      &mdwptr,
+		      &source_start, 
+		      &source_len,
+		      errstr);
 
-  //    if no svname, then we're defaulted to :_dw:
-  if (strlen (svname) == 0)
-    {
-      strcpy (svname, ":_dw:");   // strcat, strlen are safe here.
-      svnamelen = strlen (svname);
-    };
   if (internal_trace)
-    fprintf (stderr, " search-in-var: ***%s***\n", svname);
+    fprintf (stderr, 
+	     "restricted: vmidx: %ld  mdw: %ld   start: %ld  len: %ld\n",
+	     vmidx, (long) mdwptr, source_start, source_len);
+  if ( i < 0)
+    {
+      long curstmt;
+      curstmt = csl->cstmt;
+      if (i == -1)
+	nonfatalerror5 (errstr, "", CRM_ENGINE_HERE);
+      if (i == -2)
+	fatalerror5 (errstr, "", CRM_ENGINE_HERE);
+      //
+      //     did the FAULT handler change the next statement to execute?
+      //     If so, continue from there, otherwise, we FAIL.
+      if (curstmt == csl->cstmt)
+	{
+	  csl->cstmt = csl->mct[csl->cstmt]->fail_index - 1;
+	  csl->aliusstk [ csl->mct[csl->cstmt]->nest_level ] = -1;
+	};
+      goto nonfatal_route_outwards;
+    };
   
   //    get the regex pattern out of the // slashes
   crm_get_pgm_arg (pch, MAX_PATTERN, apb->s1start, apb->s1len);
   if (internal_trace)
     fprintf (stderr, " match pattern: =%s=\n", pch );
   pchlen = crm_nexpandvar (pch, apb->s1len, MAX_PATTERN);
-  
+
   if (user_trace)
     fprintf (stderr, " match pattern expands to =%s= len %ld flags %x %x \n",
 	     pch, pchlen, cflags, eflags);
@@ -251,7 +243,8 @@ int crm_expr_match (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
       long curstmt;
       curstmt = csl->cstmt;
       crm_regerror ( i, &preg, tempbuf, data_window_size);
-      fatalerror ("Regular Expression Compilation Problem:", tempbuf);
+      fatalerror5 ("Regular Expression Compilation Problem:", 
+		   tempbuf, CRM_ENGINE_HERE);
       //
       //     did the FAULT handler change the next statement to execute?
       //     If so, continue from there, otherwise, we FAIL.
@@ -263,20 +256,9 @@ int crm_expr_match (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
       goto nonfatal_route_outwards;
     };
 
-  //    get the aux match string, if one exists.
-  crm_get_pgm_arg (auxstr, MAX_PATTERN, apb->s2start, apb->s2len);
-  //   
-  //     aux match string is deprecated for TRE library newer than 0.6.3 
-  //     so we should squalk a nonfatal
-  //
-  if (apb->s2len > 0)
-    nonfatalerror ("This statement seems to use an auxillairy match string.",
-		    "Aux match strings are deprecated. ") ;
-  if (internal_trace) 
-    fprintf (stderr, " aux string pattern: =%s=\n", auxstr);
-  auxstr_len = crm_nexpandvar (auxstr, apb->s2len, MAX_PATTERN);
-  
+
   //     Get the string to be matched upon...
+#ifdef SUPERCEDED_BY_CRM_RESTRICTVAR
   vmidx = crm_vht_lookup (vht, svname, svnamelen);
   if (vht[vmidx] == NULL)
     {
@@ -289,8 +271,8 @@ int crm_expr_match (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
       //
       long curstmt;
       curstmt = csl->cstmt;
-      nonfatalerror (" Attempt to match inside nonexistent variable ( always fails!) ",
-		     svname);
+      nonfatalerror5 (" Attempt to match inside nonexistent variable ( always fails!) ",
+		      svname, CRM_ENGINE_HERE);
       //
       //     did the FAULT handler change the next statement to execute?
       //     If so, continue from there, otherwise, we FAIL.
@@ -301,6 +283,8 @@ int crm_expr_match (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 	};
       goto nonfatal_route_outwards;
     }
+#endif 
+
   //    do a check- if the vht->valtxt points into the
   //    cdw, or the tdw?
   mdw = cdw;    //  assume cdw unless otherwise proven...
@@ -311,19 +295,33 @@ int crm_expr_match (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
       && vht[vmidx]->valtxt != cdw->filetext)
     {
       long q;
-      q = fatalerror ("Bogus text block (neither cdw nor tdw) on var ",
-		svname);
+      q = fatalerror5 ("Bogus text block (neither cdw nor tdw) on var ",
+		       box_text, CRM_ENGINE_HERE);
       if (q != 0)
-	exit (EXIT_FAILURE);
+	{
+	  if (engine_exit_base != 0)
+	    {
+	      exit (engine_exit_base + 7);
+	    }
+	  else
+	    exit (EXIT_FAILURE);
+	};
     };
+#ifdef SUPERCEDED
   vtextoffset = vht[vmidx]->vstart;
   vtextend = vtextoffset + vht[vmidx]->vlen;
   vpmstart = vht[vmidx]->mstart;
   vpmend =  vpmstart + vht[vmidx]->mlen; 
+#endif
+
+  vtextoffset = source_start;
+  vtextend = source_start + source_len;
+  vpmstart = vht[vmidx]->mstart;
+  vpmend = vpmstart + vht[vmidx]->mlen;
   
   //    set up the start/end of the text we're matching against
   
-  //   default is CRM_FROMSTART 
+  //       default is CRM_FROMSTART 
   textoffset = vtextoffset;
   if (fromp == CRM_FROMSTART)
     {
@@ -353,6 +351,7 @@ int crm_expr_match (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 
   mtextlen = vtextend - textoffset;
 
+#ifdef  SUPERCEDED_BY_CRM_RESTRICTVAR
   //
   //    did the user box-specify a different start?  Combine the restrictions!
   //
@@ -367,7 +366,8 @@ int crm_expr_match (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   //    3) the earliest point we'll allow a search to go is the start of
   //       the variable + box_start
   vtextstartlimit = vtextoffset + box_start;
-
+#endif
+  vtextstartlimit = source_start;
   if (internal_trace)
     {
       fprintf (stderr, "    start matchable zone: %ld, begin search %ld, length %ld\n",
@@ -398,7 +398,7 @@ int crm_expr_match (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 	    mtextlen--;
 	    mtext = &mdw->filetext[textoffset];
 	    i = crm_regexec (&preg, mtext, mtextlen,
-			     nmatches, matches, eflags, auxstr);
+			     nmatches, matches, eflags, NULL);
 	    j = matches[0].rm_eo;
 	    if (( (textoffset + j) > oldend) && (i == 0)) done = 1;
 	  };
@@ -420,7 +420,7 @@ int crm_expr_match (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 	    mtextlen++;
 	    mtext = &mdw->filetext[textoffset];
 	    i = crm_regexec (&preg, mtext, mtextlen,
-			     nmatches, matches, eflags, auxstr);
+			     nmatches, matches, eflags, NULL);
 	    j = matches[0].rm_so;
 	  };
       };
@@ -428,7 +428,7 @@ int crm_expr_match (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
       {
 	mtext = &mdw->filetext[textoffset];
 	i = crm_regexec ( &preg, mtext, mtextlen, 
-			  nmatches, matches, eflags, auxstr);
+			  nmatches, matches, eflags, NULL);
       };
     };
   
@@ -437,10 +437,12 @@ int crm_expr_match (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   //    and now we FAIL or not...
 
 
-  if ((absentp == 1 && i != 0) || (absentp == 0 && i == 0))
+  if ((absentp == 0 && i != 0) || (absentp == 1 && i == 0))
     {
-      if (user_trace) 
-	fprintf (stderr, "Regex did not match.\n");
+      if (user_trace && !absentp) 
+	fprintf (stderr, "Regex did not match, no absent flag, failing.\n");
+      if (user_trace && absentp) 
+	fprintf (stderr, "Regex matched but with absent flag, failing.\n");
       csl->cstmt = csl->mct[csl->cstmt]->fail_index - 1;
       csl->aliusstk [ csl->mct[csl->cstmt]->nest_level ] = -1;
     }
@@ -469,13 +471,14 @@ int crm_expr_match (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
       vht[vmidx]-> mstart = matches[0].rm_so + textoffset;
       vht[vmidx]-> mlen = matches[0].rm_eo - matches[0].rm_so;
 
-      if (bindable_vars_len > 0 && absentp == 0)
-	nonfatalerror ("This program specifies an 'absent' match, and also "
+      if (bindable_vars_len > 0 && absentp )
+	nonfatalerror5 ("This program specifies an 'absent' match, and also "
 		       "tries to bind variables that, by the above, aren't "
 		       "matched!  ",
-		       "We'll ignore these variable bindings for now.");
+			"We'll ignore these variable bindings for now.",
+			CRM_ENGINE_HERE);
       
-      if ( bindable_vars_len > 0 && absentp == 1)
+      if ( bindable_vars_len > 0 && !absentp )
 	{
 	  long vstart;
 	  long vlen;
@@ -537,15 +540,18 @@ int crm_expr_match (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 		  //  will be usable.  So, we create the varname as a temp
 		  //  var, and then can reassign it with impunity.
 		  {
-		    char *vn;
+		    static char *vn;
 		    
 		    //   DANGER here - we malloc the var, use it
 		    //   in crm_set_windowed_nvar, and then free it.
 		    //   Otherwise, we'd have a memory leak.
 		    //
-		    vn = (char *) malloc (vlen + 4);
 		    if (!vn)
-		      untrappableerror("Couldn't malloc vn.\n Can't fix that.","");
+		      vn = (char *) malloc (MAX_VARNAME+16);
+		    if (!vn)
+		      untrappableerror5
+			("Couldn't malloc vn.\n Can't fix that.",
+			 "", CRM_ENGINE_HERE);
 		    strncpy (vn, &(bindable_vars[vstart]), vlen);
 		    vn[vlen] = '\000';
 		    if (strcmp (vn, ":_dw:") != 0)
@@ -582,21 +588,22 @@ int crm_expr_match (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 		      }
 		    else
 		      {
-			nonfatalerror ("This program tried to re-define the "
+			nonfatalerror5 ("This program tried to re-define the "
 				       "data window!  That's very deep and "
 				       "profound, but not acceptable.  ", 
 				       "Therefore, I'm ignoring this "
-				       "re-definition.");
+					"re-definition.", CRM_ENGINE_HERE);
 		      };
-		    free (vn);
+		    // free (vn);
 		  };
 		  //    and move on to the next binding (if any)
 		  vstart = vnext;
 		  mc++;
 		  if (mc >= MAX_SUBREGEX)
-		    nonfatalerror (
-				   "Exceeded MAX_SUBREGEX limit-too many parens in match",
-				   " Looks like you blew the gaskets on 'er.\n"); 
+		    nonfatalerror5 
+		      ( "Exceeded MAX_SUBREGEX limit-too many parens in match",
+			" Looks like you blew the gaskets on 'er.\n",
+			CRM_ENGINE_HERE); 
 		  
 		};
 	    };
@@ -617,7 +624,7 @@ int crm_expr_match (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 	  //      a var that was "non-participating", as the var will 
           //      still be in use in the VHT and thus won't be reclaimed.
 	  {
-	    int i;
+	    long i;
 	    long done = 0;
 	    long reclaimed;
 	    long maxstart, maxi;
@@ -641,10 +648,18 @@ int crm_expr_match (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 		//  reclaim that area (and no other)
 		if (maxi >= 0)
 		  {
-		    reclaimed = crm_compress_tdw_section 
+		    long j;
+		    if (internal_trace)
+		      fprintf (stderr," crm_comprss_tdw_section from match\n");
+		    // because the prev shortening, current index_starts[maxi] 
+		    // + index_lengths[maxi] may go past the end of tdw->nchars
+		    j = index_starts[maxi] + index_lengths[maxi];
+		    if (j > tdw->nchars - 1) j = tdw->nchars;
+		    reclaimed = crm_compress_tdw_section
 		      (index_texts[maxi],
 		       index_starts[maxi],
-		       index_starts[maxi] + index_lengths[maxi]);
+		       j);
+		    // WAS index_starts[maxi] + index_lengths[maxi] );
 		    if (internal_trace)
 		      fprintf (stderr,
 			       " [ MatchVar #%ld (s: %ld l: %ld) reclaimed %ld. ]\n",
@@ -660,11 +675,14 @@ int crm_expr_match (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 	  };
 	};
     };
+
+
   if (0)
     {
     nonfatal_route_outwards:
       if (user_trace)
 	fprintf (stderr, "The MATCH FAULTed and we're taking the TRAP out");
     };
+
   return (0);
 };  
