@@ -1,5 +1,5 @@
 //  crm_main.c  - Controllable Regex Mutilator,  version v1.0
-//  Copyright 2001-2004  William S. Yerazunis, all rights reserved.
+//  Copyright 2001-2006  William S. Yerazunis, all rights reserved.
 //  
 //  This software is licensed to the public under the Free Software
 //  Foundation's GNU GPL, version 2.  You may obtain a copy of the
@@ -11,7 +11,6 @@
 //
 //  include some standard files
 #include "crm114_sysincludes.h"
-
 //  include any local crm114 configuration file
 #include "crm114_config.h"
 
@@ -70,6 +69,7 @@ int main (int argc, char **argv)
   cmdline_break = -1;
   profile_execution = 0;
   prettyprint_listing = 0;
+  engine_exit_base = 0;
   q_expansion_mode = 0;
 
   //    allcate and initialize the initial root csl (control stack
@@ -78,7 +78,8 @@ int main (int argc, char **argv)
   
   csl = (CSL_CELL *) malloc (sizeof (CSL_CELL));
   if (!csl)
-    untrappableerror ("Couldn't malloc the csl.  Big problem!\n","");
+    untrappableerror5 ("Couldn't malloc the csl.  Big problem!\n",
+		       "", CRM_ENGINE_HERE);
   csl -> filename = NULL;
   csl -> filedes = -1; 
   csl -> rdwr = 0;   //  0 means readonly, 1 means read/write
@@ -97,7 +98,8 @@ int main (int argc, char **argv)
   //   and allocate the argparse block
   apb = (ARGPARSE_BLOCK *) malloc (sizeof (ARGPARSE_BLOCK));
   if (!apb)
-    untrappableerror ("Couldn't malloc apb.  This is very bad.\n","");
+    untrappableerror5 ("Couldn't malloc apb.  This is very bad.\n",
+		       "", CRM_ENGINE_HERE);
 
   //   Parse the input command arguments
   
@@ -116,7 +118,7 @@ int main (int argc, char **argv)
 	  fprintf (stderr, " CRM114 version %s (regex engine: %s)\n ", 
 		   VERSION,
 		   crm_regversion());
-	  fprintf (stderr, " Copyright 2001-2004 W.S.Yerazunis.\n");
+	  fprintf (stderr, " Copyright 2001-2006 William S. Yerazunis\n");
 	  fprintf (stderr, " This software is licensed under the GPL "
 		   "with ABSOLUTELY NO WARRANTY\n");
 	  fprintf (stderr, "     For language help, RTFRM. \n");
@@ -125,6 +127,7 @@ int main (int argc, char **argv)
 	  fprintf (stderr, " -b nn   sets a breakpoint on stmt nn\n");
 	  fprintf (stderr, " -d nn   run nn statements, then drop to debug\n");
 	  fprintf (stderr, " -e      ignore environment variables\n");
+	  fprintf (stderr, " -E      set base for engine exit values\n");
 	  fprintf (stderr, " -h      this help\n");
 	  fprintf (stderr, " -l n    listing (detail level 1 through 5)\n");
 	  fprintf (stderr, " -m nn   max number of microgroomed buckets in a chain\n");
@@ -139,10 +142,10 @@ int main (int argc, char **argv)
 	  fprintf (stderr, " -C      use env. locale (default POSIX)\n");
 	  fprintf (stderr, " -t      user trace mode on\n");
 	  fprintf (stderr, " -T      implementors trace mode on\n");
-	  fprintf (stderr, " -u dir  chdir to directory \n");
+	  fprintf (stderr, " -u dir  chdir to directory before starting\n");
 	  fprintf (stderr, " -v      print version ID and exit \n");
 	  fprintf (stderr, " -w nn   max data window size ( bytes ) \n");
-	  fprintf (stderr, " --      end of CRM114 flags; start user args\n");
+	  fprintf (stderr, " --      end of CRM114 flags; start of user args\n");
 	  fprintf (stderr, " --foo   creates var :foo: with value 'SET'\n"); 
 	  fprintf (stderr, " --x=y   creates var :x: with value 'y'\n");
 	  if (openparen > 0)
@@ -150,7 +153,12 @@ int main (int argc, char **argv)
 	      fprintf (stderr, "\n This program also claims to accept these command line args:" );
 	      fprintf (stderr, "\n  %s\n", &argv[openparen][1] );
 	    };
-	  exit (EXIT_SUCCESS);
+	  if (engine_exit_base != 0)
+	    {
+	      exit (engine_exit_base + 14);
+	    }
+	  else
+	    exit (EXIT_SUCCESS);
 	}
 
       //  -- means "end of crm114 flags" - remainder of args goes to 
@@ -298,7 +306,12 @@ int main (int argc, char **argv)
 	    {
 	      fprintf (stderr, "On -s flag: Missing or incomprehensible"
 		       ".CSS file length.\n");
-	      exit (EXIT_FAILURE);
+	      if (engine_exit_base != 0)
+		{
+		  exit (engine_exit_base + 15);
+		}
+	      else
+		exit (EXIT_FAILURE);
 	    };
 	  
 	  if (user_trace > 0)
@@ -318,6 +331,18 @@ int main (int argc, char **argv)
 		     cmdline_break);
 	  goto end_command_line_parse_loop;
 	};
+
+      //   set base value for detailed engine exit values
+      if (strncmp (argv[i], "-E", 2) == 0 && strlen(argv[i]) == 2)
+	{
+	  i++;    // move to the next arg
+	  if (i < argc)
+	    sscanf (argv[i], "%ld", &engine_exit_base);
+	  if (user_trace > 0)
+	    fprintf (stderr, "Setting the engine exit base value to %ld\n", 
+		     engine_exit_base);
+	  goto end_command_line_parse_loop;
+	};
       
       //   set countdown cycles before dropping to debugger
       if (strncmp (argv[i], "-d", 2) == 0 && strlen(argv[i]) == 2)
@@ -363,12 +388,19 @@ int main (int argc, char **argv)
       
       if (strncmp (argv[i], "-v", 2) == 0 && strlen(argv[i]) == 2)
 	{
-	  fprintf (stderr, " This is CRM114, version %s (%s)\n", 
+	  //   NOTE - version info goes to stdout, not stderr, just like GCC does
+	  fprintf (stdout, 
+		   " This is CRM114, version %s (%s)\n", 
 		   VERSION, 
 		   crm_regversion());
-	  fprintf (stderr, " Copyright 2001-2004 W.S.Yerazunis.\n");
-	  fprintf (stderr, " This software is licensed under the GPL with ABSOLUTELY NO WARRANTY\n");
-	  exit( EXIT_SUCCESS );
+	  fprintf (stdout, " Copyright 2001-2006 William S. Yerazunis\n");
+	  fprintf (stdout, " This software is licensed under the GPL with ABSOLUTELY NO WARRANTY\n");
+	  if (engine_exit_base != 0)
+	    {
+	      exit (engine_exit_base + 16);
+	    }
+	  else
+	    exit( EXIT_SUCCESS );
 	};
       
       if (strncmp (argv[i], "-{", 2) == 0)  //  don't care about the "}"
@@ -451,7 +483,8 @@ int main (int argc, char **argv)
       if (csl->filename == NULL)
 	{
           if (strlen(argv[i]) > MAX_FILE_NAME_LEN)
-	    untrappableerror ("Invalid filename, ", "filename too long.");
+	    untrappableerror5 ("Invalid filename, ", 
+			       "filename too long.", CRM_ENGINE_HERE);
 	  csl->filename = argv[i];
 	  if (user_trace > 0)
 	    fprintf (stderr, "Using program file %s\n", csl->filename);
@@ -462,6 +495,7 @@ int main (int argc, char **argv)
 		 i);
     }
   
+  //  main2 ();
   
   //
   //     Did we get a program filename?  If not, look for one.
@@ -475,8 +509,8 @@ int main (int argc, char **argv)
 	if (argv[i][0] != '-')
 	  {
            if (strlen(argv[i]) > MAX_FILE_NAME_LEN)
-	     untrappableerror ("Couldn't open the file, ",
-			       "filename too long.");
+	     untrappableerror5 ("Couldn't open the file, ",
+				"filename too long.", CRM_ENGINE_HERE);
 	   csl->filename = argv[i];
 	   i = argc;
 	  }
@@ -490,7 +524,12 @@ int main (int argc, char **argv)
       fprintf (stderr, "\nCan't find a file to run,"
 	       "or a command-line to execute. \n"
 	       "I give up... (exiting)\n");
-      exit (EXIT_SUCCESS);
+      if (engine_exit_base != 0)
+	{
+	  exit (engine_exit_base + 17);
+	}
+      else
+	exit (EXIT_SUCCESS);
     };
   
   //     open, stat and load the program file
@@ -501,7 +540,12 @@ int main (int argc, char **argv)
 	  {
 	    fprintf (stderr, "CRM114 version %s \n", VERSION);
 	    fprintf (stderr, "Try 'crm <progname>', or 'crm -h' for help\n");
-	    exit (EXIT_SUCCESS);
+	    if (engine_exit_base != 0)
+	      {
+		exit (engine_exit_base + 18);
+	      }
+	    else
+	      exit (EXIT_SUCCESS);
 	  }
 	else
 	  {
@@ -517,12 +561,15 @@ int main (int argc, char **argv)
       //   if we got here, then it's a command-line program, and
       //   we should just assemble the proggie from the argv [openbracket] 
       if (strlen (&(argv[openbracket][1])) + 2048 > max_pgmsize)
-	untrappableerror ("The command line programis too big. \n",
-			  "Try increasing the max program size with -P. \n");
+	untrappableerror5 ("The command line program is too big. \n",
+			   "Try increasing the max program size with -P. \n",
+			   CRM_ENGINE_HERE);
       csl->filename = "(from command line)";
       csl->filetext = (char *) malloc (sizeof (char) * max_pgmsize);
       if (!csl->filetext)
-	untrappableerror ("Couldn't malloc csl->filetext space (where I was going to put your program.\nWithout program space, we can't run.  Sorry.","");
+	untrappableerror5 
+	  ("Couldn't malloc csl->filetext space (where I was going to put your program.\nWithout program space, we can't run.  Sorry.",
+	   "", CRM_ENGINE_HERE);
       strcpy (csl->filetext, "\n");
       //     the [1] below gets rid of the leading - sign
       strcat (csl->filetext, &(argv[openbracket][1]));
@@ -542,13 +589,14 @@ int main (int argc, char **argv)
   
   cdw = malloc (sizeof (CSL_CELL));
   if (!cdw)
-    untrappableerror ("Couldn't malloc cdw.\nThis is very bad.","");
+    untrappableerror5 ("Couldn't malloc cdw.\nThis is very bad.","", 
+		       CRM_ENGINE_HERE);
   cdw->filename = NULL;
   cdw->rdwr = 1;
   cdw->filedes = -1;
   cdw->filetext = malloc (sizeof (char) * data_window_size);
   if (!cdw->filetext)
-    untrappableerror ("Couldn't malloc cdw->filetext.\nWithout this space, you have no place for data.  Thus, we cannot run.","");
+    untrappableerror5 ("Couldn't malloc cdw->filetext.\nWithout this space, you have no place for data.  Thus, we cannot run.","", CRM_ENGINE_HERE);
   //      also allocate storage for the windowed data input
   newinputbuf = malloc (sizeof (char) * data_window_size);
   
@@ -559,24 +607,28 @@ int main (int argc, char **argv)
   outbuf = malloc (sizeof (char) * data_window_size);
   tempbuf = malloc (sizeof (char) * data_window_size);
   if (!tempbuf || !outbuf || !inbuf || !newinputbuf)
-    untrappableerror (
+    untrappableerror5 (
 		      "Couldn't malloc one or more of"
 		      "newinputbuf,inbuf,outbuf,tempbuf.\n"
 		      "These are all necessary for operation."
-                      "We can't run.","");
+                      "We can't run.","", CRM_ENGINE_HERE);
 
   //     Initialize the VHT, add in a few predefined variables
   //
   crm_vht_init(argc, argv);
   
   //    Call the pre-processor on the program
-  
+  //
   status = crm_preprocessor (csl, 0);
-  
+
   //    Now, call the microcompiler on the program file.
   status = crm_microcompiler ( csl, vht);
   //    Great - program file is now mapped via csl->mct
   
+  //    Put a copy of the preprocessor-result text into 
+  //    the isolated variable ":_pgm_text:"
+  crm_set_temp_var (":_pgm_text:", csl->filetext);
+
   //  If the windowflag == 0, we should preload the data window.  Now,
   //  let's get some data in.  
   
@@ -605,7 +657,11 @@ int main (int argc, char **argv)
       int readsize = data_window_size - 1;
       while (! feof (stdin) && i < data_window_size - 1)
 	{
-	  i += fread (cdw->filetext + i, 1, readsize-1, stdin);
+	  //i += fread (cdw->filetext + i, 1, readsize-1, stdin);
+	  int rs;
+	  rs = i + readsize < data_window_size - 1 ? 
+	    readsize : data_window_size - i - 1;
+	  i+= fread (cdw->filetext + i, 1, rs, stdin);
 	  if (feof (stdin))
 	    {
 	      break;
@@ -669,7 +725,8 @@ int main (int argc, char **argv)
 		 cdw->filetext,
 		 0,
 		 cdw->nchars,
-		 -1);
+		 -1,
+		 0);
   };
   //
   //    We also set up the :_iso: to hold the isolated variables.
@@ -700,7 +757,8 @@ int main (int argc, char **argv)
 		 tdw->filetext,
 		 0,
 		 0,
-		 -1);
+		 -1,
+		 0);
   };
 #endif
   //    Now we're here, we can actually run!
@@ -709,7 +767,8 @@ int main (int argc, char **argv)
   
   status = crm_invoke ();
   
+  //     This is the *real* exit from the engine, so we do not override
+  // the engine's exit status with an engine_exit_base value.
   exit ( (char) status);
-  
  }
 
