@@ -1,5 +1,5 @@
 //  crm_markovian.c  - Controllable Regex Mutilator,  version v1.0
-//  Copyright 2001-2004  William S. Yerazunis, all rights reserved.
+//  Copyright 2001-2007 William S. Yerazunis, all rights reserved.
 //  
 //  This software is licensed to the public under the Free Software
 //  Foundation's GNU GPL, version 2.  You may
@@ -57,7 +57,8 @@ static long hctable[] =
 
 //    How to learn Markovian style.
 //
-int crm_expr_markov_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
+int crm_expr_markov_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb,
+			   char *txtptr, long txtstart, long txtlen)
 {
   //     learn the sparse spectrum of this input window as 
   //     belonging to a particular type.
@@ -82,9 +83,11 @@ int crm_expr_markov_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   long textoffset;
   long textmaxoffset;
   long sense;
-  long vhtindex;
   long microgroom;
+  long max_feature_terms;     //  how many terms do we actually get per word?
   long fev;
+  long learns_index;
+  long features_index;
 
   //    seen_features - for UNIQUEness tests
   long unique_mode;
@@ -121,7 +124,7 @@ int crm_expr_markov_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   seen_features = NULL;
   if (apb->sflags & CRM_NOCASE)
     {
-      cflags = cflags || REG_ICASE;
+      cflags = cflags | REG_ICASE;
       eflags = 1;
       if (user_trace)
 	fprintf (stderr, "turning oncase-insensitive match\n");
@@ -146,6 +149,15 @@ int crm_expr_markov_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
       if (user_trace)
 	fprintf (stderr, " enabling microgrooming.\n");
     };
+
+  //     How many features in a Markovian?  Here's the answer:
+  max_feature_terms = ( 1 << (MARKOVIAN_WINDOW_LEN - 1 ));
+  if (apb->sflags & CRM_UNIGRAM)
+    {
+      max_feature_terms = 1;
+      if (user_trace)
+	fprintf (stderr, " enabling unigram-only (Bayesian) features.\n");
+    };
   //
   //             grab the filename, and stat the file
   //      note that neither "stat", "fopen", nor "open" are
@@ -174,7 +186,12 @@ int crm_expr_markov_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 	  fprintf (stderr, 
 		"\n Couldn't open your new CSS file %s for writing; errno=%d .\n",
 		 &htext[i], errno);
-	  exit (EXIT_FAILURE);
+	  if (engine_exit_base != 0)
+	    {
+	      exit (engine_exit_base + 19);
+	    }
+	  else
+	    exit (EXIT_FAILURE);
         };
       //       do we have a user-specified file size?
       if (sparse_spectrum_file_length == 0 ) {
@@ -202,36 +219,130 @@ int crm_expr_markov_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   //
   //         mmap the hash file into memory so we can bitwhack it
   //
-  hashes = (FEATUREBUCKET_TYPE *) crm_mmap_file (NULL, &(htext[i]), hfsize); 
+  hashes = (FEATUREBUCKET_TYPE *) crm_mmap_file (&(htext[i]), 
+						 0, hfsize, 
+						 PROT_READ | PROT_WRITE, 
+						 MAP_SHARED,
+						 NULL); 
   if (hashes == MAP_FAILED)
     {
-      fev = fatalerror ("Couldn't get access to the statistics file named: ",
-			&htext[i]);
+      fev = fatalerror5 ("Couldn't get access to the statistics file named: ",
+			 &htext[i], CRM_ENGINE_HERE);
       return (fev);
     };
 
   //
   //        check the version of the file
   //
-  if (hashes[0].hash != 0 ||
-      hashes[0].key  != 0 ||
-      hashes[0].value!= 0)
-    {
-      fev =fatalerror ("The .css file is the wrong version!  Filename is: ",
-		       &htext[i]);
-      return (fev);
-    };
+  //  if (hashes[0].hash != 0 ||
+  //    hashes[0].key  != 0 ||
+  //    hashes[0].value!= 0)
+  //  {
+  //    fev =fatalerror ("The .css file is the wrong version!  Filename is: ",
+  //		       &htext[i]);
+  //    return (fev);
+  //  };
+
   //
   //   now set the hfsize to the number of entries, not the number
   //   of bytes total
   hfsize = hfsize / sizeof ( FEATUREBUCKET_TYPE );
 
+
+#ifdef OSB_LEARNCOUNTS
+  //       Reserve the OSB_LEARNCOUNT buckets, even though Markov doesn't
+  //       use them.
+  //
+  //       We use the reserved h2 == 0 setup for the learncount.
+  //  
+  {
+    char* litf = "Learnings in this file";
+    char* fitf = "Features in this file";
+    unsigned long hcode, h1, h2;
+    //
+    hcode = strnhash (litf, strlen ( litf ));
+    h1 = hcode % hfsize;
+    h2 = 0;
+    if (hashes[h1].hash != hcode)
+      {
+	// initialize the file?
+	if (hashes[h1].hash == 0 && hashes[h1].key == 0) 
+	  {
+	    hashes[h1].hash = hcode;
+	    hashes[h1].key = 0;
+	    hashes[h1].value = 1;
+	    learns_index = h1;
+	  }
+	else
+	  { 
+	    //fatalerror (" This file should have learncounts, but doesn't!",
+	    //	" The slot is busy, too.  It's hosed.  Time to die.");
+	    // goto regcomp_failed;
+	  };
+      }
+    else 
+      {
+	if (hashes[h1].key == 0)
+	  //   the learncount matched.
+	  {
+	    learns_index = h1;
+	    if (sense > 0)
+	      hashes[h1].value = hashes[h1].value + sense;
+	    else 
+	      {
+	      if (hashes[h1].value + sense > 0)
+		hashes[h1].value += sense;
+	      else
+		hashes[h1].value = 0;
+	      }
+	    if (user_trace)
+	      fprintf (stderr, "This file has had %ld documents learned!\n",
+		       hashes[h1].value);
+	  };
+      };
+    hcode = strnhash (fitf, strlen ( fitf ));
+    h1 = hcode % hfsize;
+    h2 = 0;
+    if (hashes[h1].hash != hcode)
+      {
+	// initialize the file?
+	if (hashes[h1].hash == 0 && hashes[h1].key == 0) 
+	  {
+	    hashes[h1].hash = hcode;
+	    hashes[h1].key = 0;
+	    hashes[h1].value = 1;
+	    features_index = h1;
+	  }
+	else
+	  { 
+	    //fatalerror (" This file should have learncounts, but doesn't!",
+	    //		" The slot is busy, too.  It's hosed.  Time to die.");
+	    //goto regcomp_failed ;
+	  };
+      }
+    else 
+      {
+	if (hashes[h1].key == 0)
+	  //   the learncount matched.
+	  {
+	    features_index = h1;
+	    if (user_trace)
+	      fprintf (stderr, "This file has had %ld features learned!\n",
+		       hashes[h1].value);
+	  };
+      };
+  };
+
+#endif 
+  
+
   if (unique_mode)
     {
       seen_features = calloc (hfsize, 1);
       if (seen_features == NULL)
-        untrappableerror (" Couldn't allocate enough memory to keep track",
-                          "of nonunique features.  This is deadly");
+        untrappableerror5 (" Couldn't allocate enough memory to keep track",
+			   "of nonunique features.  This is deadly", 
+			   CRM_ENGINE_HERE);
     };
 
 
@@ -243,7 +354,8 @@ int crm_expr_markov_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   if ( i > 0)
     {
       crm_regerror ( i, &regcb, tempbuf, data_window_size);
-      nonfatalerror ("Regular Expression Compilation Problem:", tempbuf);
+      nonfatalerror5 ("Regular Expression Compilation Problem:", 
+		      tempbuf, CRM_ENGINE_HERE);
       goto regcomp_failed;
     };
   
@@ -254,6 +366,9 @@ int crm_expr_markov_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   k = 0;
   j = 0;
   i = 0;
+
+#ifdef OLD_STUPID_VAR_RESTRICTION
+
   if (llen > 0)
     {
       vhtindex = crm_vht_lookup (vht, ltext, llen);
@@ -266,8 +381,8 @@ int crm_expr_markov_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   if (vht[vhtindex] == NULL)
     {
       long q;
-      q = fatalerror (" Attempt to LEARN from a nonexistent variable ",
-		  ltext);
+      q = fatalerror5 (" Attempt to LEARN from a nonexistent variable ",
+		       ltext, CRM_ENGINE_HERE);
       return (q);
     };
   mdw = NULL;
@@ -278,12 +393,19 @@ int crm_expr_markov_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   if (mdw == NULL)
     {
       long q;
-      q = fatalerror (" Bogus text block containing variable ", ltext);  
+      q = fatalerror5 (" Bogus text block containing variable ", 
+		       ltext, CRM_ENGINE_HERE);  
       return (q);
     }
   textoffset = vht[vhtindex]->vstart;
   textmaxoffset = textoffset + vht[vhtindex]->vlen;
-  
+#endif  
+
+  //  No need to do any parsing of a box restriciton, as those
+  //   are passed in to us from the caller.
+  textoffset = txtstart;
+  textmaxoffset = txtstart + txtlen;
+
   //   init the hashpipe with 0xDEADBEEF 
   for (h = 0; h < MARKOVIAN_WINDOW_LEN; h++)
     {
@@ -296,33 +418,34 @@ int crm_expr_markov_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
     {
       long wlen;
       long slen;
-      unsigned char *ptok = &(mdw->filetext[textoffset]);
-      unsigned char *ptok_max = &(mdw->filetext[textmaxoffset]);
 
       //  do the regex
       //  slen = endpoint (= start + len) 
       //        - startpoint (= curr textoffset)
-      slen = vht[vhtindex]->vstart + vht[vhtindex]->vlen - textoffset ;
+      //      slen = txtlen;
+      slen = textmaxoffset - textoffset;
 
       // if pattern is empty, extract non graph delimited tokens
       // directly ([[graph]]+) instead of calling regexec (8% faster!)
       if (ptext[0] != '\0')
 	{     
-	  k = crm_regexec ( &regcb, &(mdw->filetext[textoffset]),
+	  k = crm_regexec ( &regcb, &(txtptr[textoffset]),
 			    slen, 
 			    5, match, 0, NULL);
 	}
       else
 	{
 	  k = 0;
-	  while ((ptok < ptok_max) && !isgraph ((int) *ptok))
-	    ptok++;
-	  match[0].rm_so = ptok - (unsigned char *) &(mdw->filetext[textoffset]);
-	  while ((ptok < ptok_max) && isgraph ((int) *ptok))
-	    ptok++;
-	  match[0].rm_eo = ptok - (unsigned char *) &(mdw->filetext[textoffset]);
-	  wlen = match[0].rm_eo - match[0].rm_so;
-	  if (wlen <= 0 || ptok > ptok_max)
+	  //         skip non-graphical characthers
+	  match[0].rm_so = 0;
+	  while (!isgraph (txtptr[textoffset + match[0].rm_so])
+		 && textoffset + match[0].rm_so < textmaxoffset)
+	    match[0].rm_so ++;
+	  match[0].rm_eo = match[0].rm_so;
+	  while (isgraph (txtptr [textoffset + match[0].rm_eo])
+		 && textoffset + match[0].rm_eo < textmaxoffset)
+	    match[0].rm_eo ++;
+	  if ( match[0].rm_so == match[0].rm_eo)
 	    k = 1;
 	}
       
@@ -332,7 +455,7 @@ int crm_expr_markov_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
       {
 	wlen = match[0].rm_eo - match[0].rm_so;
 	memmove (tempbuf, 
-		 &(mdw->filetext[textoffset + match[0].rm_so]),
+		 &(txtptr[textoffset + match[0].rm_so]),
 		 wlen);
 	tempbuf[wlen] = '\000';
 	
@@ -349,8 +472,9 @@ int crm_expr_markov_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 	  };
 	if (match[0].rm_eo == 0)
 	  {
-	    nonfatalerror ( "The LEARN pattern matched zero length! ",
-			    "\n Forcing an increment to avoid an infinite loop.");
+	    nonfatalerror5 ( "The LEARN pattern matched zero length! ",
+			     "\n Forcing an increment to avoid an infinite loop.",
+			     CRM_ENGINE_HERE);
 	    match[0].rm_eo = 1;
 	  };
 
@@ -395,24 +519,8 @@ int crm_expr_markov_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 	    //
 	    th = 0;
 	    //	    for ( j = 0; j <= 15 ; j++)
-	    for (j = 0; j < (1 << (MARKOVIAN_WINDOW_LEN - 1)); j++)
+	    for (j = 0; j < max_feature_terms; j++)
 	      {
-#ifdef FOO
-		//
-		//   First Order Only - 1 token per.
-		//
-		hindex = hashpipe [0] ;
-		if (hindex == 0) hindex = 1;
-		h1 = hindex;
-		hindex = hindex % hfsize;
-		if (hindex == 0) hindex = 1;
-		
-		//   this is the secondary (crosscut) hash, used for
-		//   confirmation of the key value.
-		h2 =   hashpipe[0];
-		if (h2 == 0) h2 = 0xdeadbeef;
-		j = 16;
-#endif
 #ifdef TGB
 		//
 		//   Token Grab Bag - ignore sequence, distance. alias
@@ -500,7 +608,6 @@ int crm_expr_markov_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 		};
 	      if (h2 == 0) h2 = 0xdeadbeef;
 #endif
-
 #ifdef SBPH
 		hindex = hashpipe [0] 
 		  + (  3 * hashpipe[1] * ((j>>0) & 0x0001))
@@ -595,13 +702,13 @@ int crm_expr_markov_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 		    //      can hold no more features (this is unrecoverable)
 		    if (incrs > hfsize - 3)
 		      {
-			nonfatalerror ("Your program is stuffing too many "
+			nonfatalerror5 ("Your program is stuffing too many "
 				       "features into this size .css file.  "
 				       "Adding any more features is "
 				       "impossible in this file.",
 				       "You are advised to build a larger "
 				       ".css file and merge your data into "
-				       "it.");
+					"it.", CRM_ENGINE_HERE);
 			goto learn_end_regex_loop;
 		      };
 		    hindex++;
@@ -654,8 +761,9 @@ int crm_expr_markov_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
  regcomp_failed:
 
   
-  //  and remember to let go of the mmap 
-  crm_munmap_file ( (void *) hashes );
+  //  and remember to let go of all the mmaps (full flush) 
+  //  crm_munmap_all ();
+  crm_force_munmap_addr (hashes);
 
   //   and let go of the seen_features array
   if (seen_features) free (seen_features);
@@ -687,7 +795,8 @@ int crm_expr_markov_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 
 //      How to Markovian CLASSIFY some text.
 //
-int crm_expr_markov_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
+int crm_expr_markov_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb,
+			      char *txtptr, long txtstart, long txtlen)
 {
   //      classify the sparse spectrum of this input window 
   //      as belonging to a particular type.
@@ -699,13 +808,13 @@ int crm_expr_markov_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   long h;          //  we use h for our hashpipe counter, as needed.
   char ptext[MAX_PATTERN];  //  the regex pattern
   long plen;
-  char ltext[MAX_PATTERN];  //  the variable to classify
-  long llen;
+  //  char ltext[MAX_PATTERN];  //  the variable to classify
+  //long llen;
   //  the hash file names
   char htext[MAX_PATTERN+MAX_CLASSIFIERS*MAX_FILE_NAME_LEN];  
   long htext_maxlen = MAX_PATTERN+MAX_CLASSIFIERS*MAX_FILE_NAME_LEN;
   long hlen;
-  //  the match statistics variable inbuf
+  //  the match statistics variable 
   char stext[MAX_PATTERN+MAX_CLASSIFIERS*(MAX_FILE_NAME_LEN+100)]; 
   long stext_maxlen = MAX_PATTERN+MAX_CLASSIFIERS*(MAX_FILE_NAME_LEN+100); 
   long slen;
@@ -717,7 +826,8 @@ int crm_expr_markov_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   long cflags;
   long not_microgroom;     //  is microgrooming disabled (fast-quit 
                            //   optimization if 0th feature gone)
-  long vhtindex;
+  long max_feature_terms;   //  how many features do we get at each
+                            //    pipe position?
 
   struct stat statbuf;      //  for statting the hash file
                             //  longest association set in the hashing
@@ -735,7 +845,7 @@ int crm_expr_markov_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   double htf;             // hits this feature got.
   double tprob;         //  total probability in the "success" domain.
 
-  double textlen;    //  text length  - rougly corresponds to
+  //  double textlen;    //  text length  - rougly corresponds to
                         //  information content of the text to classify
   
   double ptc[MAX_CLASSIFIERS]; // current running probability of this class
@@ -769,11 +879,11 @@ int crm_expr_markov_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   if (internal_trace)
     fprintf (stderr, "executing a CLASSIFY\n");
   
-  //           extract the variable name (if present)
+  //          We get the variable block, start, and len from caller
   //
-  crm_get_pgm_arg (ltext, MAX_PATTERN, apb->b1start, apb->b1len);
-  llen = apb->b1len;
-  llen = crm_nexpandvar (ltext, llen, MAX_PATTERN);
+  // crm_get_pgm_arg (ltext, MAX_PATTERN, apb->b1start, apb->b1len);
+  // llen = apb->b1len;
+  // llen = crm_nexpandvar (ltext, llen, MAX_PATTERN);
   
   //           extract the hash file names
   crm_get_pgm_arg (htext, htext_maxlen, apb->p1start, apb->p1len);
@@ -825,6 +935,14 @@ int crm_expr_markov_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 	fprintf (stderr, " disabling fast-skip optimization.\n");
     };
 
+  max_feature_terms = ( 1 << (MARKOVIAN_WINDOW_LEN - 1 ));
+  if (apb->sflags & CRM_UNIGRAM)
+    {
+      max_feature_terms = 1;
+      if (user_trace)
+	fprintf (stderr, " enabling unigram-only (Bayesian) features.\n");
+    };
+
   unique_mode = 0;
   if (apb->sflags & CRM_UNIQUE)
     {
@@ -841,7 +959,8 @@ int crm_expr_markov_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   if ( i > 0)
     {
       crm_regerror ( i, &regcb, tempbuf, data_window_size);
-      nonfatalerror ("Regular Expression Compilation Problem:", tempbuf);
+      nonfatalerror5 ("Regular Expression Compilation Problem:", 
+		      tempbuf, CRM_ENGINE_HERE);
       goto regcomp_failed;
     };
   
@@ -966,8 +1085,8 @@ int crm_expr_markov_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 	    {
 	      if (vbar_seen)
 		{
-		  nonfatalerror ("Only one ' | ' allowed in a CLASSIFY. \n" ,
-				 "We'll ignore it for now.");
+		  nonfatalerror5 ("Only one ' | ' allowed in a CLASSIFY. \n" ,
+				  "We'll ignore it for now.", CRM_ENGINE_HERE);
 		}
 	      else
 		{
@@ -983,8 +1102,8 @@ int crm_expr_markov_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 	      //             quick check- does the file even exist?
 	      if (k != 0)
 		{
-		  nonfatalerror ("Nonexistent Classify table named: ",
-				 fname);
+		  nonfatalerror5 ("Nonexistent Classify table named: ",
+				  fname, CRM_ENGINE_HERE);
 		}
 	      else
 		{
@@ -994,14 +1113,17 @@ int crm_expr_markov_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 		  hashlens[maxhash] = statbuf.st_size;
 		  //  mmap the hash file into memory so we can bitwhack it
 		  hashes[maxhash] = (FEATUREBUCKET_TYPE *) crm_mmap_file (
-						   NULL, 
-						   fname, 
-						   hashlens[maxhash]);
+						   fname,
+						   0, hashlens[maxhash],
+						   PROT_READ | PROT_WRITE,
+						   MAP_SHARED,
+						   NULL);
+						   
 		  if (hashes[maxhash] == MAP_FAILED )
 		    {
-		      nonfatalerror ("Couldn't get access to the "
+		      nonfatalerror5 ("Couldn't get access to the "
 				     "statistics file named: ",
-				     fname);
+				      fname, CRM_ENGINE_HERE);
 		    }
 		  else
 		    {
@@ -1009,24 +1131,27 @@ int crm_expr_markov_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 		      //
 		      //     Check to see if this file is the right version
 		      //
-		      long fev;
-		      if (hashes[maxhash][0].hash != 0 ||
-			  hashes[maxhash][0].key  != 0 ||
-			  hashes[maxhash][0].value!= 0)
-			{
-			  fev = fatalerror ("The .css file is the wrong version!  Filename is: ",
-					   fname);
-			  return (fev);
-			};
+		      //long fev;
+		      //if (hashes[maxhash][0].hash != 0 ||
+		      //	  hashes[maxhash][0].key  != 0 ||
+		      //  hashes[maxhash][0].value!= 0)
+		      //{
+		      //  fev = fatalerror ("The .css file is the wrong version!  Filename is: ",
+		      //		   fname);
+		      //  return (fev);
+		      //};
 		      
+
 		      //  set this hashlens to the length in
 		      //  features instead of the length in bytes.
 		      
 		      hashlens[maxhash] = hashlens[maxhash] / sizeof (FEATUREBUCKET_TYPE);
 		      hashname[maxhash] = (char *) malloc (fnlen+10);
 		      if (!hashname[maxhash])
-			untrappableerror(
-					 "Couldn't malloc hashname[maxhash]\n","We need that part later, so we're stuck.  Sorry.");
+			untrappableerror5
+			  ("Couldn't malloc hashname[maxhash]\n",
+			  "We need that part later, so we're stuck.  Sorry.", 
+			  CRM_ENGINE_HERE);
 		      strncpy(hashname[maxhash],fname,fnlen);
 		      hashname[maxhash][fnlen]='\000';
 		      maxhash++;
@@ -1035,8 +1160,9 @@ int crm_expr_markov_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 	    };
 	
 	  if (maxhash > MAX_CLASSIFIERS-1)
-	    nonfatalerror ("Too many classifier files.",
-			   "Some may have been disregarded");
+	    nonfatalerror5 ("Too many classifier files.",
+			    "Some may have been disregarded",
+			    CRM_ENGINE_HERE);
 	};
     };
 
@@ -1060,7 +1186,8 @@ int crm_expr_markov_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   //	do we have at least 1 valid .css files?
   if (maxhash == 0)
     {
-      fatalerror ("Couldn't open at least 2 .css files for classify().", "");
+      nonfatalerror5 ("Couldn't open at least 1 .css files for classify().", 
+		   "", CRM_ENGINE_HERE);
     };
   //	do we have at least 1 valid .css file at both sides of '|'?
   //if (!vbar_seen || succhash < 0 || (maxhash < succhash + 2))
@@ -1081,9 +1208,9 @@ int crm_expr_markov_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 	    //     Note that we _calloc_, not malloc, to zero the memory first.
 	    seen_features[ifile] = calloc (hashlens[ifile]+1, 1);
 	    if ( seen_features[ifile] == NULL )
-	      untrappableerror (" Couldn't allocate enough memory to keep " 
+	      untrappableerror5 (" Couldn't allocate enough memory to keep " 
 				" track of nonuniuqe features.  ",
-				"This is deadly. ");
+				 "This is deadly. ", CRM_ENGINE_HERE);
 	  }
       else
         seen_features[ifile] = NULL;
@@ -1129,6 +1256,7 @@ int crm_expr_markov_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   k = 0;
   thistotal = 0;
 
+#ifdef OLD_STUPID_VAR_RESTRICTION
   if (llen > 0)
     {
       vhtindex = crm_vht_lookup (vht, ltext, llen );
@@ -1139,8 +1267,8 @@ int crm_expr_markov_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
     }
   if (vht[vhtindex] == NULL)
     {
-      return (fatalerror (" Attempt to CLASSIFY from a nonexistent variable ",
-			  ltext));
+      return (fatalerror5 (" Attempt to CLASSIFY from a nonexistent variable ",
+			   ltext, CRM_ENGINE_HERE));
     };
   mdw = NULL;
   if (tdw->filetext == vht[vhtindex]->valtxt)
@@ -1148,12 +1276,17 @@ int crm_expr_markov_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   if (cdw->filetext == vht[vhtindex]->valtxt)
     mdw = cdw;
   if (mdw == NULL)
-    return ( fatalerror (" Bogus text block containing variable ", ltext));  
+    return ( fatalerror5 (" Bogus text block containing variable ", 
+			  ltext, CRM_ENGINE_HERE));  
   textoffset = vht[vhtindex]->vstart;
   textmaxoffset = textoffset + vht[vhtindex]->vlen;
 
   textlen = (vht[vhtindex]->vlen);
   if (textlen < 1.0) textlen = 1.0; 
+#endif
+
+  textoffset = txtstart;
+  textmaxoffset = txtstart + txtlen;
 
   //   init the hashpipe with 0xDEADBEEF 
   for (h = 0; h < MARKOVIAN_WINDOW_LEN; h++)
@@ -1170,30 +1303,33 @@ int crm_expr_markov_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
     {
       long wlen;
       long slen;
-      unsigned char *ptok = &(mdw->filetext[textoffset]);
-      unsigned char *ptok_max = &(mdw->filetext[textmaxoffset]);
+      //unsigned char *ptok = &(mdw->filetext[textoffset]);
+      //unsigned char *ptok_max = &(mdw->filetext[textmaxoffset]);
 
       //  do the regex  
-      slen = vht[vhtindex]->vstart + vht[vhtindex]->vlen - textoffset ;
+      //      slen = vht[vhtindex]->vstart + vht[vhtindex]->vlen - textoffset ;
+      //slen = txtlen;
+      slen = textmaxoffset - textoffset;
 
       // if pattern is empty, extract non graph delimited tokens
       // directly ([[graph]]+) instead of calling regexec  (8% faster!)
       if (ptext[0] != '\0')
 	{
-	  k = crm_regexec (&regcb, &(mdw->filetext[textoffset]),
+	  k = crm_regexec (&regcb, &(txtptr[textoffset]),
 			   slen, 5, match, 0, NULL);
 	}
       else
 	{
 	  k = 0;
-	  while ((ptok < ptok_max) && !isgraph ((int) *ptok))
-	    ptok++;
-	  match[0].rm_so = ptok - (unsigned char *) &(mdw->filetext[textoffset]);
-	  while ((ptok < ptok_max) && isgraph ((int) *ptok))
-	    ptok++;
-	  match[0].rm_eo = ptok - (unsigned char *) &(mdw->filetext[textoffset]);
-	  wlen = match[0].rm_eo - match[0].rm_so;
-	  if (wlen <= 0 || ptok > ptok_max)
+	  match[0].rm_so = 0;
+	  while (!isgraph (txtptr[textoffset + match[0].rm_so])
+		 && textoffset + match[0].rm_so < textmaxoffset)
+	    match[0].rm_so ++;
+	  match[0].rm_eo = match[0].rm_so;
+	  while (isgraph (txtptr [textoffset + match[0].rm_eo])
+		 && textoffset + match[0].rm_eo < textmaxoffset)
+	    match[0].rm_eo ++;
+	  if ( match[0].rm_so == match[0].rm_eo)
 	    k = 1;
 	}
 
@@ -1202,14 +1338,15 @@ int crm_expr_markov_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
       
       wlen = match[0].rm_eo - match[0].rm_so;
       memmove (tempbuf, 
-	       &(mdw->filetext[textoffset + match[0].rm_so]),
+	       &(txtptr[textoffset + match[0].rm_so]),
 	       wlen);
       tempbuf[wlen] = '\000';
       
       if (match[0].rm_eo == 0)
 	{
-	  nonfatalerror("The CLASSIFY pattern matched zero length! ",
-			"\n Forcing an increment to avoid an infinite loop.");
+	  nonfatalerror5 ("The CLASSIFY pattern matched zero length! ",
+			  "\n Forcing an increment to avoid an infinite loop.",
+			  CRM_ENGINE_HERE);
 	  match[0].rm_eo = 1;
 	};
       if (internal_trace)
@@ -1265,7 +1402,7 @@ int crm_expr_markov_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 	  th = 0;
 	  //	  for ( j = 0; j < 16 ; j++)
 	  for (j = 0; 
-	       j < (1 << (MARKOVIAN_WINDOW_LEN - 1)); 
+	       j < max_feature_terms; 
 	       j++)
 	    {
 #ifdef FOO
@@ -1903,8 +2040,15 @@ int crm_expr_markov_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
     {
       //      close (hfds [k]);
       crm_munmap_file ( (void *) hashes[k]);
+
+      //   and let go of the seen_features array
+      if (seen_features [k]) free (seen_features[k]);
+      seen_features[k] = NULL;
+
     };
 
+
+
   //   and one last chance to force probabilities into the non-stuck zone
   //
   //  if (pic == 0.0 ) pic = DBL_MIN;
@@ -1946,7 +2090,7 @@ int crm_expr_markov_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
       //  There would be a possible buffer overflow except that _we_ control
       //   what gets written here.  So it's no biggie.
 
-      if (tprob >= 0.5000)
+      if (tprob > 0.5000)
 	{
 	  sprintf (buf, "CLASSIFY succeeds; success probability: %6.4f  pR: %6.4f\n", tprob, overall_pR );
 	}
@@ -2003,10 +2147,11 @@ int crm_expr_markov_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
       // whining on stderr
       if (strcmp(&(stext[strlen(stext)-strlen(buf)]), buf) != 0)
         {
-          nonfatalerror( "WARNING: not enough room in the buffer to create "
+          nonfatalerror5( "WARNING: not enough room in the buffer to create "
 			 "the statistics text.  Perhaps you could try bigger "
 			 "values for MAX_CLASSIFIERS or MAX_FILE_NAME_LEN?",
-			 " ");
+			  " ",
+			  CRM_ENGINE_HERE);
 	};
       crm_destructive_alter_nvariable (svrbl, svlen, 
 				       stext, strlen (stext));
@@ -2017,7 +2162,7 @@ int crm_expr_markov_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   //
   for (i = 0; i < maxhash; i++)
     free (hashname[i]);
-  if (tprob < 0.5000)
+  if (tprob <= 0.5000)
     {
       if (user_trace)
 	fprintf (stderr, "CLASSIFY was a FAIL, skipping forward.\n");
