@@ -1,5 +1,5 @@
 //  crm_osb_winnow.c  - Controllable Regex Mutilator,  version v1.0
-//  Copyright 2001-2004  William S. Yerazunis, all rights reserved.
+//  Copyright 2001-2006  William S. Yerazunis, all rights reserved.
 //  
 //  This software is licensed to the public under the Free Software
 //  Foundation's GNU GPL, version 2.  You may obtain a copy of the
@@ -64,7 +64,8 @@ static long spectra_start;
 //    be set to 0 in the evaluation.
 //
 
-int crm_expr_osb_winnow_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
+int crm_expr_osb_winnow_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb,
+			       char *txtptr, long txtstart, long txtlen)
 {
   //     learn the osb_winnow transform spectrum of this input window as 
   //     belonging to a particular type.
@@ -74,8 +75,6 @@ int crm_expr_osb_winnow_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   long h;                   //  h is our counter in the hashpipe;
   char ptext[MAX_PATTERN];  //  the regex pattern
   long plen;
-  char ltext[MAX_PATTERN];  //  the variable to learn
-  long llen;
   char htext[MAX_PATTERN];  //  the hash name
   long hlen;
   long cflags, eflags;
@@ -91,8 +90,8 @@ int crm_expr_osb_winnow_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   long textoffset;
   long textmaxoffset;
   float sense;
-  long vhtindex;
   long microgroom;
+  long use_unigrams;
   long fev;
   long made_new_file; 
   
@@ -109,10 +108,10 @@ int crm_expr_osb_winnow_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   hlen = apb->p1len;
   hlen = crm_nexpandvar (htext, hlen, MAX_PATTERN);
   //
-  //           extract the variable name (if present)
-  crm_get_pgm_arg (ltext, MAX_PATTERN, apb->b1start, apb->b1len);
-  llen = apb->b1len;
-  llen = crm_nexpandvar (ltext, llen, MAX_PATTERN);
+  //        We get the varname and var-restriction from the caller now
+  //  crm_get_pgm_arg (ltext, MAX_PATTERN, apb->b1start, apb->b1len);
+  // llen = apb->b1len;
+  // llen = crm_nexpandvar (ltext, llen, MAX_PATTERN);
   
   //     get the "this is a word" regex
   crm_get_pgm_arg (ptext, MAX_PATTERN, apb->s1start, apb->s1len);
@@ -127,7 +126,7 @@ int crm_expr_osb_winnow_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 
   if (apb->sflags & CRM_NOCASE)
     {
-      cflags = cflags || REG_ICASE;
+      cflags = cflags | REG_ICASE;
       eflags = 1;
       if (user_trace)
 	fprintf (stderr, "turning oncase-insensitive match\n");
@@ -147,6 +146,7 @@ int crm_expr_osb_winnow_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
       if (user_trace)
 	fprintf (stderr, " refuting learning\n");
     };
+
   microgroom = 0;
   if (apb->sflags & CRM_MICROGROOM)
     {
@@ -155,6 +155,15 @@ int crm_expr_osb_winnow_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 	fprintf (stderr, " enabling microgrooming.\n");
     };
 
+  use_unigrams = 0;
+  if (apb->sflags & CRM_UNIGRAM)
+    {
+      use_unigrams = 1;
+      if (user_trace)
+	fprintf (stderr, " enabling unigram-only operation.\n");
+    };
+
+
   //
   //             grab the filename, and stat the file
   //      note that neither "stat", "fopen", nor "open" are
@@ -179,14 +188,19 @@ int crm_expr_osb_winnow_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
       //      file didn't exist... create it
       FILE *f;
       if (user_trace)
-	fprintf (stderr, "\nHad to create new COW file %s\n", fname);
+	fprintf (stderr, "\n Opening new COW file %s for write\n", fname);
       f = fopen (fname, "wb");
       if (!f)
 	{
 	  fprintf (stderr, 
 		"\n Couldn't open your new COW file %s for writing; errno=%d .\n",
 		 fname, errno);
-	  exit (EXIT_FAILURE);
+	  if (engine_exit_base != 0)
+	    {
+	      exit (engine_exit_base + 21);
+	    }
+	  else
+	    exit (EXIT_FAILURE);
         };
       //       do we have a user-specified file size?
       if (sparse_spectrum_file_length == 0 ) {
@@ -215,13 +229,17 @@ int crm_expr_osb_winnow_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   //
   //         open the .cow hash file into memory so we can bitwhack it
   //
-  hashes = (WINNOW_FEATUREBUCKET_STRUCT *) crm_mmap_file ( NULL,
-							   fname,
-							   hfsize);
+  hashes = (WINNOW_FEATUREBUCKET_STRUCT *) 
+    crm_mmap_file ( 
+		   fname,
+		   0, hfsize,
+		   PROT_READ | PROT_WRITE,
+		   MAP_SHARED,
+		   NULL);
   if (hashes == MAP_FAILED)
     {
-      fev = fatalerror ("Couldn't memory-map the .cow file named: ",
-			fname);
+      fev = fatalerror5 ("Couldn't memory-map the .cow file named: ",
+			 fname, CRM_ENGINE_HERE);
       return (fev);
     };
 
@@ -242,7 +260,7 @@ int crm_expr_osb_winnow_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
       fprintf (stderr, "Hash was: %ld, key was %ld\n", hashes[0].hash, hashes[0].key);
       fev =fatalerror ("The .cow file is the wrong type!  We're expecting "
 		       "a Osb_Winnow-spectrum file.  The filename is: ",
-		       fname);
+		       fname, CRM_ENGINE_HERE);
       return (fev);
     };
 #endif
@@ -264,9 +282,9 @@ int crm_expr_osb_winnow_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   //
   xhashes = calloc ( hfsize, (sizeof (unsigned char)) );
   if ( !xhashes )
-    untrappableerror(
+    untrappableerror5(
 		     "Couldn't malloc xhashes\n",
-		     "We need that part.  Sorry.\n");
+		     "We need that part.  Sorry.\n", CRM_ENGINE_HERE);
 
   //   compile the word regex
   //
@@ -276,7 +294,8 @@ int crm_expr_osb_winnow_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   if ( i > 0)
     {
       crm_regerror ( i, &regcb, tempbuf, data_window_size);
-      nonfatalerror ("Regular Expression Compilation Problem:", tempbuf);
+      nonfatalerror5 ("Regular Expression Compilation Problem:", 
+		      tempbuf, CRM_ENGINE_HERE);
       goto regcomp_failed;
     };
   
@@ -287,6 +306,8 @@ int crm_expr_osb_winnow_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   k = 0;
   j = 0;
   i = 0;
+
+#ifdef OLD_STUPID_VAR_RESTRICTION
   if (llen > 0)
     {
       vhtindex = crm_vht_lookup (vht, ltext, llen);
@@ -299,8 +320,8 @@ int crm_expr_osb_winnow_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   if (vht[vhtindex] == NULL)
     {
       long q;
-      q = fatalerror (" Attempt to LEARN from a nonexistent variable ",
-		  ltext);
+      q = fatalerror5 (" Attempt to LEARN from a nonexistent variable ",
+		       ltext, CRM_ENGINE_HERE);
       return (q);
     };
   mdw = NULL;
@@ -311,12 +332,18 @@ int crm_expr_osb_winnow_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   if (mdw == NULL)
     {
       long q;
-      q = fatalerror (" Bogus text block containing variable ", ltext);  
+      q = fatalerror5 (" Bogus text block containing variable ", 
+		       ltext, CRM_ENGINE_HERE);  
       return (q);
     }
   textoffset = vht[vhtindex]->vstart;
   textmaxoffset = textoffset + vht[vhtindex]->vlen;
-  
+#endif
+
+  textoffset = txtstart;
+  textmaxoffset = txtstart + txtlen;
+
+
   //   init the hashpipe with 0xDEADBEEF 
   for (h = 0; h < OSB_WINNOW_WINDOW_LEN; h++)
     {
@@ -329,32 +356,35 @@ int crm_expr_osb_winnow_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
     {
       long wlen;
       long slen;
-      unsigned char *ptok = &(mdw->filetext[textoffset]);
-      unsigned char *ptok_max = &(mdw->filetext[textmaxoffset]);
+      //      unsigned char *ptok = &(mdw->filetext[textoffset]);
+      //unsigned char *ptok_max = &(mdw->filetext[textmaxoffset]);
 
       //  do the regex
       //  slen = endpoint (= start + len) 
       //        - startpoint (= curr textoffset)
-      slen = vht[vhtindex]->vstart + vht[vhtindex]->vlen - textoffset ;
+      //      slen = txtlen ;
+      slen = textmaxoffset - textoffset;
 
       // if pattern is empty, extract non graph delimited tokens
       // directly ([[graph]]+) instead of calling regexec
       if (ptext[0] != '\0')
 	{
-	  k = crm_regexec (&regcb, &(mdw->filetext[textoffset]),
+	  k = crm_regexec (&regcb, &(txtptr[textoffset]),
 			   slen, 5, match, 0, NULL);
 	}
       else
 	{
 	  k = 0;
-	  while ((ptok < ptok_max) && !isgraph ((int) *ptok))
-	    ptok++;
-	  match[0].rm_so = ptok - (unsigned char *) &(mdw->filetext[textoffset]);
-	  while ((ptok < ptok_max) && isgraph ((int) *ptok))
-	    ptok++;
-	  match[0].rm_eo = ptok - (unsigned char *) &(mdw->filetext[textoffset]);
-	  wlen = match[0].rm_eo - match[0].rm_so;
-	  if (wlen <= 0 || ptok > ptok_max)
+	  //         skip non-graphical characthers
+	  match[0].rm_so = 0;
+	  while (!isgraph (txtptr[textoffset + match[0].rm_so])
+		 && textoffset + match[0].rm_so < textmaxoffset)
+	    match[0].rm_so ++;
+	  match[0].rm_eo = match[0].rm_so;
+	  while (isgraph (txtptr [textoffset + match[0].rm_eo])
+		 && textoffset + match[0].rm_eo < textmaxoffset)
+	    match[0].rm_eo ++;
+	  if ( match[0].rm_so == match[0].rm_eo)
 	    k = 1;
 	}
       
@@ -364,7 +394,7 @@ int crm_expr_osb_winnow_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
       {
 	wlen = match[0].rm_eo - match[0].rm_so;
 	memmove (tempbuf, 
-		 &(mdw->filetext[textoffset + match[0].rm_so]),
+		 &(txtptr[textoffset + match[0].rm_so]),
 		 wlen);
 	tempbuf[wlen] = '\000';
 	
@@ -381,8 +411,10 @@ int crm_expr_osb_winnow_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 	  };
 	if (match[0].rm_eo == 0)
 	  {
-	    nonfatalerror ( "The LEARN pattern matched zero length! ",
-			    "\n Forcing an increment to avoid an infinite loop.");
+	    nonfatalerror5 
+	      ( "The LEARN pattern matched zero length! ",
+		"\n Forcing an increment to avoid an infinite loop.",
+		CRM_ENGINE_HERE);
 	    match[0].rm_eo = 1;
 	  };
 
@@ -431,12 +463,23 @@ int crm_expr_osb_winnow_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 		 j < OSB_WINNOW_WINDOW_LEN;
 		 j++)
 	      {
-		h1 = hashpipe[0]*hctable[0] + hashpipe[j] * hctable[j<<1];
-		if (h1 < spectra_start)
-		  h1 = spectra_start;
-		h2 = hashpipe[0]*hctable[1] + hashpipe[j] * hctable[(j<<1)-1];
-		if (h2 == 0) h2 = 0xdeadbeef;
-
+		if (use_unigrams)
+		  {
+		    h1 = hashpipe[0]*hctable[0];
+		    if (h1 < spectra_start)
+		      h1 = spectra_start;
+		    h2 = hashpipe[0]*hctable[1];
+		    if (h2 == 0) h2 = 0xdeadbeef;
+		    j = OSB_WINNOW_WINDOW_LEN;
+		  }
+		else
+		  {
+		    h1 = hashpipe[0]*hctable[0] + hashpipe[j] * hctable[j<<1];
+		    if (h1 < spectra_start)
+		      h1 = spectra_start;
+		    h2 = hashpipe[0]*hctable[1] + hashpipe[j] * hctable[(j<<1)-1];
+		    if (h2 == 0) h2 = 0xdeadbeef;
+		  };
 		hindex = h1 % hfsize;
 		if (hindex < spectra_start ) hindex = spectra_start;
 
@@ -493,13 +536,13 @@ int crm_expr_osb_winnow_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 		    //      can hold no more features (this is unrecoverable)
 		    if (incrs > hfsize - 3)
 		      {
-			nonfatalerror ("Your program is stuffing too many "
+			nonfatalerror5 ("Your program is stuffing too many "
 				       "features into this size .cow file.  "
 				       "Adding any more features is "
 				       "impossible in this file.",
 				       "You are advised to build a larger "
 				       ".cow file and merge your data into "
-				       "it.");
+					"it.", CRM_ENGINE_HERE);
 			goto learn_end_regex_loop;
 		      };
 		    //
@@ -553,6 +596,8 @@ int crm_expr_osb_winnow_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
  regcomp_failed:
 
   //  and remember to let go of the mmap and the pattern bufffer
+  // (and force a cache purge)
+  // crm_munmap_all ();
   crm_munmap_file ((void *) hashes);
 
   free (xhashes);
@@ -565,6 +610,7 @@ int crm_expr_osb_winnow_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   //    The fix is to do a trivial read/write on the .cow file, to force
   //    the filesystem to repropagate it's caches.
   //
+
   {
     int hfd;                  //  hashfile fd
     FEATURE_HEADER_STRUCT foo;
@@ -584,7 +630,8 @@ int crm_expr_osb_winnow_learn (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 
 //      How to do a Osb_Winnow CLASSIFY some text.
 //
-int crm_expr_osb_winnow_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
+int crm_expr_osb_winnow_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb,
+				  char *txtptr, long txtstart, long txtlen)
 {
   //      classify the sparse spectrum of this input window 
   //      as belonging to a particular type.
@@ -596,8 +643,6 @@ int crm_expr_osb_winnow_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   long h;          //  we use h for our hashpipe counter, as needed.
   char ptext[MAX_PATTERN];  //  the regex pattern
   long plen;
-  char ltext[MAX_PATTERN];  //  the variable to classify
-  long llen;
   //  the hash file names
   char htext[MAX_PATTERN+MAX_CLASSIFIERS*MAX_FILE_NAME_LEN];  
   long htext_maxlen = MAX_PATTERN+MAX_CLASSIFIERS*MAX_FILE_NAME_LEN;
@@ -612,8 +657,8 @@ int crm_expr_osb_winnow_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   char fname[MAX_FILE_NAME_LEN];
   long eflags;
   long cflags;
-  long vhtindex;
   long not_microgroom = 1;
+  long use_unigrams;
 
   struct stat statbuf;      //  for statting the hash file
   unsigned long hashpipe[OSB_WINNOW_WINDOW_LEN+1]; 
@@ -633,7 +678,7 @@ int crm_expr_osb_winnow_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   double htf;             // hits this feature got.
   double tprob = 0;         //  total probability in the "success" domain.
 
-  double textlen;    //  text length  - rougly corresponds to
+  //double textlen;    //  text length  - rougly corresponds to
                         //  information content of the text to classify
 
   WINNOW_FEATUREBUCKET_STRUCT *hashes[MAX_CLASSIFIERS];
@@ -658,11 +703,12 @@ int crm_expr_osb_winnow_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   if (internal_trace)
     fprintf (stderr, "executing an OSB-WINNOW CLASSIFY\n");
   
-  //           extract the variable name (if present)
+  // 
+  //      We get the to-be-classified text from the caller now.
   //
-  crm_get_pgm_arg (ltext, MAX_PATTERN, apb->b1start, apb->b1len);
-  llen = apb->b1len;
-  llen = crm_nexpandvar (ltext, llen, MAX_PATTERN);
+  //  crm_get_pgm_arg (ltext, MAX_PATTERN, apb->b1start, apb->b1len);
+  // llen = apb->b1len;
+  // llen = crm_nexpandvar (ltext, llen, MAX_PATTERN);
   
   //           extract the hash file names
   crm_get_pgm_arg (htext, htext_maxlen, apb->p1start, apb->p1len);
@@ -712,6 +758,14 @@ int crm_expr_osb_winnow_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 	fprintf (stderr, " disabling fast-skip optimization.\n");
     };
 
+  use_unigrams = 0;
+  if (apb->sflags & CRM_UNIGRAM)
+    {
+      use_unigrams = 1;
+      if (user_trace)
+	fprintf (stderr, " enabling unigram-only operation.\n");
+    };
+
   //   compile the word regex
   if ( internal_trace)
     fprintf (stderr, "\nWordmatch pattern is %s", ptext);
@@ -719,7 +773,8 @@ int crm_expr_osb_winnow_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   if ( i > 0)
     {
       crm_regerror ( i, &regcb, tempbuf, data_window_size);
-      nonfatalerror ("Regular Expression Compilation Problem:", tempbuf);
+      nonfatalerror5 ("Regular Expression Compilation Problem:", 
+		      tempbuf, CRM_ENGINE_HERE);
       goto regcomp_failed;
     };
   
@@ -801,8 +856,8 @@ int crm_expr_osb_winnow_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 	    {
 	      if (vbar_seen)
 		{
-		  nonfatalerror ("Only one ' | ' allowed in a CLASSIFY. \n" ,
-				 "We'll ignore it for now.");
+		  nonfatalerror5 ("Only one ' | ' allowed in a CLASSIFY. \n" ,
+				  "We'll ignore it for now.", CRM_ENGINE_HERE);
 		}
 	      else
 		{
@@ -818,8 +873,8 @@ int crm_expr_osb_winnow_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 	      //             quick check- does the file even exist?
 	      if (k != 0)
 		{
-		  nonfatalerror ("Nonexistent Classify table named: ",
-				 fname);
+		  nonfatalerror5 ("Nonexistent Classify table named: ",
+				  fname, CRM_ENGINE_HERE);
 		}
 	      else
 		{
@@ -828,12 +883,16 @@ int crm_expr_osb_winnow_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 		  hashlens[maxhash] = statbuf.st_size;
 		  //  mmap the hash file into memory so we can bitwhack it
 		  hashes[maxhash] = (WINNOW_FEATUREBUCKET_STRUCT *)
-		    crm_mmap_file ( 0, fname, hashlens[maxhash]);
+		    crm_mmap_file ( fname, 
+				    0, hashlens[maxhash],
+				    PROT_READ,
+				    MAP_SHARED,
+				    NULL);
 
 		  if (hashes[maxhash] == MAP_FAILED )
 		    {
-		      nonfatalerror ("Couldn't memory-map the table file",
-				     fname);
+		      nonfatalerror5 ("Couldn't memory-map the table file",
+				      fname, CRM_ENGINE_HERE);
 		    }
 		  else
 		    {
@@ -845,9 +904,10 @@ int crm_expr_osb_winnow_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 		      if (hashes[maxhash][0].hash != 1 ||
 			  hashes[maxhash][0].key  != 0) 
 			{
-			  fev =fatalerror ("The .css file is the wrong type!  We're expecting "
-					   "a Osb_Winnow-spectrum file.  The filename is: ",
-					   &htext[i]);
+			  fev =fatalerror5 
+			    ("The .css file is the wrong type!  We're expecting "
+			     "a Osb_Winnow-spectrum file.  The filename is: ",
+			     &htext[i], CRM_ENGINE_HERE);
 			  return (fev);
 			};
 #endif
@@ -861,8 +921,8 @@ int crm_expr_osb_winnow_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 		      hashlens[maxhash] = hashlens[maxhash] / sizeof (WINNOW_FEATUREBUCKET_STRUCT);
 		      hashname[maxhash] = (char *) malloc (fnlen+10);
 		      if (!hashname[maxhash])
-			untrappableerror(
-					 "Couldn't malloc hashname[maxhash]\n","We need that part later, so we're stuck.  Sorry.");
+			untrappableerror5 (
+					   "Couldn't malloc hashname[maxhash]\n","We need that part later, so we're stuck.  Sorry.", CRM_ENGINE_HERE);
 		      strncpy(hashname[maxhash],fname,fnlen);
 		      hashname[maxhash][fnlen]='\000';
 		      
@@ -872,17 +932,19 @@ int crm_expr_osb_winnow_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 		      xhashes[maxhash] = calloc (hashlens[maxhash], 
 						 sizeof (unsigned char));
 		      if (!xhashes[maxhash])
-			untrappableerror(
+			untrappableerror5(
 					 "Couldn't malloc xhashes[maxhash]\n",
-					 "We need that part.  Sorry.\n");
+					 "We need that part.  Sorry.\n",
+					  CRM_ENGINE_HERE);
 		      
 		      maxhash++;
 		    };
 		};
 	    };
 	  if (maxhash > MAX_CLASSIFIERS-1)
-	    nonfatalerror ("Too many classifier files.",
-			   "Some may have been disregarded");
+	    nonfatalerror5 ("Too many classifier files.",
+			    "Some may have been disregarded",
+			    CRM_ENGINE_HERE);
 	};
     };
   
@@ -904,7 +966,8 @@ int crm_expr_osb_winnow_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   //	do we have at least 1 valid .css files?
   if (maxhash == 0)
     {
-      fatalerror ("Couldn't open at least 1 .cow files for classify().", "");
+      fatalerror5 ("Couldn't open at least 1 .cow files for classify().", 
+		   "", CRM_ENGINE_HERE);
     };
   //	do we have at least 1 valid .cow file at both sides of '|'?
   //if (!vbar_seen || succhash < 0 || (maxhash < succhash + 2))
@@ -947,6 +1010,7 @@ int crm_expr_osb_winnow_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   k = 0;
   thistotal = 0;
 
+#ifdef OLD_STUPID_VAR_RESTRICTION
   if (llen > 0)
     {
       vhtindex = crm_vht_lookup (vht, ltext, llen );
@@ -957,8 +1021,8 @@ int crm_expr_osb_winnow_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
     }
   if (vht[vhtindex] == NULL)
     {
-      return (fatalerror (" Attempt to CLASSIFY from a nonexistent variable ",
-			  ltext));
+      return (fatalerror5 (" Attempt to CLASSIFY from a nonexistent variable ",
+			   ltext, CRM_ENGINE_HERE));
     };
   mdw = NULL;
   if (tdw->filetext == vht[vhtindex]->valtxt)
@@ -966,12 +1030,17 @@ int crm_expr_osb_winnow_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   if (cdw->filetext == vht[vhtindex]->valtxt)
     mdw = cdw;
   if (mdw == NULL)
-    return ( fatalerror (" Bogus text block containing variable ", ltext));  
+    return ( fatalerror5 (" Bogus text block containing variable ", 
+			 ltext, CRM_ENGINE_HERE));  
   textoffset = vht[vhtindex]->vstart;
   textmaxoffset = textoffset + vht[vhtindex]->vlen;
 
   textlen = (vht[vhtindex]->vlen);
   if (textlen < 1.0) textlen = 1.0; 
+#endif
+  textoffset = txtstart;
+  textmaxoffset = txtstart + txtlen;
+
 
   //   init the hashpipe with 0xDEADBEEF 
   for (h = 0; h < OSB_WINNOW_WINDOW_LEN; h++)
@@ -988,30 +1057,33 @@ int crm_expr_osb_winnow_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
     {
       long wlen;
       long slen;
-      unsigned char *ptok = &(mdw->filetext[textoffset]);
-      unsigned char *ptok_max = &(mdw->filetext[textmaxoffset]);
+      //      unsigned char *ptok = &(mdw->filetext[textoffset]);
+      //  unsigned char *ptok_max = &(mdw->filetext[textmaxoffset]);
 
       //  do the regex  
-      slen = vht[vhtindex]->vstart + vht[vhtindex]->vlen - textoffset ;
+      //      slen = txtlen - textoffset;
+      slen = textmaxoffset - textoffset;
 
       // if pattern is empty, extract non graph delimited tokens
       // directly ([[graph]]+) instead of calling regexec
       if (ptext[0] != '\0')
 	{
-	  k = crm_regexec (&regcb, &(mdw->filetext[textoffset]),
+	  k = crm_regexec (&regcb, &(txtptr[textoffset]),
 			   slen, 5, match, 0, NULL);
 	}
       else
 	{
 	  k = 0;
-	  while ((ptok < ptok_max) && !isgraph ((int) *ptok))
-	    ptok++;
-	  match[0].rm_so = ptok - (unsigned char *) &(mdw->filetext[textoffset]);
-	  while ((ptok < ptok_max) && isgraph ((int) *ptok))
-	    ptok++;
-	  match[0].rm_eo = ptok - (unsigned char *) &(mdw->filetext[textoffset]);
-	  wlen = match[0].rm_eo - match[0].rm_so;
-	  if (wlen <= 0 || ptok > ptok_max)
+	  //         skip non-graphical characthers
+	  match[0].rm_so = 0;
+	  while (!isgraph (txtptr[textoffset + match[0].rm_so])
+		 && textoffset + match[0].rm_so < textmaxoffset)
+	    match[0].rm_so ++;
+	  match[0].rm_eo = match[0].rm_so;
+	  while (isgraph (txtptr [textoffset + match[0].rm_eo])
+		 && textoffset + match[0].rm_eo < textmaxoffset)
+	    match[0].rm_eo ++;
+	  if ( match[0].rm_so == match[0].rm_eo)
 	    k = 1;
 	}
 
@@ -1020,16 +1092,10 @@ int crm_expr_osb_winnow_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
       
       wlen = match[0].rm_eo - match[0].rm_so;
       memmove (tempbuf, 
-	       &(mdw->filetext[textoffset + match[0].rm_so]),
+	       &(txtptr[textoffset + match[0].rm_so]),
 	       wlen);
       tempbuf[wlen] = '\000';
       
-      if (match[0].rm_eo == 0)
-	{
-	  nonfatalerror("The CLASSIFY pattern matched zero length! ",
-			"\n Forcing an increment to avoid an infinite loop.");
-	  match[0].rm_eo = 1;
-	};
       if (internal_trace)
 	{		   
 	  fprintf (stderr, 
@@ -1044,8 +1110,10 @@ int crm_expr_osb_winnow_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 	};
       if (match[0].rm_eo == 0)
 	{
-	  nonfatalerror ( "The CLASSIFYpattern matched zero length! ",
-			  "\n Forcing an increment to avoid an infinite loop.");
+	  nonfatalerror5 
+	    ( "The CLASSIFY pattern matched zero length! ",
+	      "\n Forcing an increment to avoid an infinite loop.",
+	      CRM_ENGINE_HERE);
 	  match[0].rm_eo = 1;
 	};
       //  slide previous hashes up 1
@@ -1090,17 +1158,25 @@ int crm_expr_osb_winnow_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
 	       j < OSB_WINNOW_WINDOW_LEN;
 	       j++)
 	    {
-	      //h1 = hashpipe[0]*hctable[0] 
-	      //	+ (j != 0) * hashpipe[j] * hctable[1<<j];
-	      h1 = hashpipe[0]*hctable[0] + hashpipe[j] * hctable[j<<1];
-	      if (h1 < spectra_start)
-		h1 = spectra_start;
-	      //h2 = hashpipe[0]*hctable[1] 
-	      //		+ (j != 0) * hashpipe[j] * hctable[(1<<j)+1];
-	      h2 = hashpipe[0]*hctable[1] + hashpipe[j] * hctable[(j<<1)-1];
-	      if (h2 == 0) h2 = 0xdeadbeef;
-	      hindex = h1;
+	      if (use_unigrams)
+		{
+		  h1 = hashpipe[0]*hctable[0];
+		  if (h1 < spectra_start)
+		    h1 = spectra_start;
+		  h2 = hashpipe[0]*hctable[1];
+		  if (h2 == 0) h2 = 0xdeadbeef;
+		  j = OSB_WINNOW_WINDOW_LEN;
+		}
+	      else
+		{
+		  h1 = hashpipe[0]*hctable[0] + hashpipe[j] * hctable[j<<1];
+		  if (h1 < spectra_start)
+		    h1 = spectra_start;
+		  h2 = hashpipe[0]*hctable[1] + hashpipe[j] * hctable[(j<<1)-1];
+		  if (h2 == 0) h2 = 0xdeadbeef;
+		};
 
+	      hindex = h1;
 	      if (internal_trace)
 		fprintf (stderr, "Polynomial %d has h1:%ld  h2: %ld\n",
 			 j, h1, h2);
@@ -1243,7 +1319,7 @@ int crm_expr_osb_winnow_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
       //  There would be a possible buffer overflow except that _we_ control
       //   what gets written here.  So it's no biggie.
 
-      if (tprob >= 0.5000)
+      if (tprob > 0.5000)
 	{
 	  sprintf (buf, "CLASSIFY succeeds; success probability: %6.4f  pR: %6.4f\n", tprob, overall_pR );
 	}
@@ -1310,10 +1386,11 @@ int crm_expr_osb_winnow_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
       // whining on stderr
       if (strcmp(&(stext[strlen(stext)-strlen(buf)]), buf) != 0)
         {
-          nonfatalerror( "WARNING: not enough room in the buffer to create "
-			 "the statistics text.  Perhaps you could try bigger "
-			 "values for MAX_CLASSIFIERS or MAX_FILE_NAME_LEN?",
-			 " ");
+          nonfatalerror5 
+	    ( "WARNING: not enough room in the buffer to create "
+	      "the statistics text.  Perhaps you could try bigger "
+	      "values for MAX_CLASSIFIERS or MAX_FILE_NAME_LEN?",
+	      " ", CRM_ENGINE_HERE);
 	};
       if (svlen > 0)
 	crm_destructive_alter_nvariable (svrbl, svlen, 
@@ -1325,7 +1402,7 @@ int crm_expr_osb_winnow_classify (CSL_CELL *csl, ARGPARSE_BLOCK *apb)
   //
   for (i = 0; i < maxhash; i++)
     free (hashname[i]);
-  if (tprob < 0.5000)
+  if (tprob <= 0.5000)
     {
       if (user_trace)
 	fprintf (stderr, "CLASSIFY was a FAIL, skipping forward.\n");
