@@ -1,5 +1,5 @@
 //  crm_var_hash_table.c  - Controllable Regex Mutilator,  version v1.0
-//  Copyright 2001-2004  William S. Yerazunis, all rights reserved.
+//  Copyright 2001-2006  William S. Yerazunis, all rights reserved.
 //  
 //  This software is licensed to the public under the Free Software
 //  Foundation's GNU GPL, version 2.  You may obtain a copy of the
@@ -41,7 +41,7 @@ extern char *tempbuf;
 void crm_vht_init (int argc, char **argv)
 {
   long i, j, k;
-  long uvstart = 0;
+  long uvstart = 0;   // uvstart is the arg that the user sees (post "--")
   long uvlist = 0;
   char uvset[MAX_VARNAME];
   extern char **environ;
@@ -51,8 +51,9 @@ void crm_vht_init (int argc, char **argv)
  //   create the variable hash table (one big one, shared )
   vht = (VHT_CELL **) malloc (sizeof (VHT_CELL *) * vht_size);
   if (!vht)
-    untrappableerror("Couldn't malloc VHT cell.\n",
-		     "No VHT cells, no variables, so no can run.  Sorry.");
+    untrappableerror5("Couldn't malloc VHT cell.\n",
+		      "No VHT cells, no variables, so no can run.  Sorry.",
+		      CRM_ENGINE_HERE);
   for (i = 0; i < vht_size; i++)
     vht[i] = NULL;
 
@@ -60,17 +61,18 @@ void crm_vht_init (int argc, char **argv)
   //    initialize the temporary (non-data-window) area...
   tdw = malloc (sizeof (CSL_CELL));
   if (!tdw)
-    untrappableerror("Couldn't malloc tdw.\n"  
+    untrappableerror5 ("Couldn't malloc tdw.\n"  
 		     "We need the TDW for isolated variables."  
-		     "Can't continue.  Sorry.\n","");
+		       "Can't continue.  Sorry.\n","", CRM_ENGINE_HERE);
   tdw->filename = NULL;
   tdw->rdwr = 1;
   tdw->filedes = -1;
   tdw->filetext = malloc (sizeof (char) * data_window_size);
   if (!tdw->filetext)
-      untrappableerror("Couldn't malloc tdw->filetext.\n"
+      untrappableerror5("Couldn't malloc tdw->filetext.\n"
 		       "Without this space, you can't have any isolated "
-		       "variables,\n and we're stuck.  Sorry.","");
+			"variables,\n and we're stuck.  Sorry.","",
+			CRM_ENGINE_HERE);
   tdw->filetext[0] = '\000';
   tdw->nchars = 0;
   tdw->hash = 0;
@@ -87,6 +89,9 @@ void crm_vht_init (int argc, char **argv)
   crm_set_temp_var (":_sl:", "/");
   crm_set_temp_var (":_sc:", ";");
   crm_set_temp_var (":_cd:", "0");
+  crm_set_temp_var ("::", " ");
+
+  //   put the version string in as a variable.
   {
     char verstr[1025];
     verstr[0] = 0;
@@ -100,30 +105,45 @@ void crm_vht_init (int argc, char **argv)
   //  
   //    install the argc and argv values; restart argv values from [2]
   //    if a "--" metaflag is seen.
+  //
+  //    argv[0] and argv[1] are not overrideable by "--".
+  crm_set_temp_var ( ":_arg0:", argv[0] );
+  crm_set_temp_var ( ":_arg1:", argv[1] );
+
+  //     Check to see if there's a "--" arg.  If so, mark uvstart
+  //   (that is, "user var start" at that point)... but only the first "--".
+  {
+    long i, j;
+    uvstart = 2;
+    i = 0; j = 0;
+    for (i = 2; argc > i; i++)
+      {
+	//   Check for the "--" metaflag
+	if (strlen (argv[i]) == 2 
+	    && strncmp (argv[i], "--", 2) == 0
+	    && uvstart == 2)
+	  {
+	    if (internal_trace)
+	      fprintf (stderr, "Resetting uvstart counter to 2\n");
+	    uvstart = i+1;
+	  };
+      }
+  };
+
+  //       The user variables start at argv[uvstart]
   {
     long i, j;
     char anamebuf [255];
-    i = 0; j = 0; 
-    for ( i = 0; argc > i; i++ )
+    j = 2;
+    for ( i = uvstart; argc > i; i++ )
       {
-	//   check for the "--" metaflag
-	if (strlen ( argv[i] ) == 2 && strncmp (argv[i], "--", 2) == 0)
-	  {
-	    if (internal_trace) 
-	      fprintf (stderr, "Resetting _arg counter to 2\n");
-	    j = 2;
-	    if (uvstart == 0) uvstart = i;
-	  }
-	else
-	  {
-	    sprintf (anamebuf, ":_arg%ld:", j);
-	    crm_set_temp_var ( anamebuf, argv[i] );
-	    j++;
-	  };
+	sprintf (anamebuf, ":_arg%ld:", j);
+	crm_set_temp_var ( anamebuf, argv[i] );
+	j++;
       };
     //
     //    and put the "user-visible" argc into a var as well.
-    sprintf (anamebuf, "%ld", j);
+    sprintf (anamebuf, "%ld", j );
     crm_set_temp_var (":_argc:", anamebuf);
     //
     //   Go through argv, and place positional arguments (that is,
@@ -159,6 +179,28 @@ void crm_vht_init (int argc, char **argv)
     //   and set the fault to be a null string for now.
     crm_set_temp_var (":_fault:", "");
     //
+    //   set the current line number to a set of zeroes...
+    crm_set_temp_var (":_cs:", "00000000");
+    //
+    //   Set the "lazy" intermediate variable to just a space.
+    //   This will get rebound to point to the active lazy var.
+    crm_set_temp_var (":_lazy:", " ");
+
+    //    set the current pid and parent pid.
+    {
+      char pidstr [32];
+      long pid;
+      pid = (long) getpid();
+      sprintf (pidstr, "%ld", pid);
+      crm_set_temp_var (":_pid:", pidstr);
+#ifdef POSIX
+      pid = (long) getppid();
+      sprintf (pidstr, "%ld", pid);
+      crm_set_temp_var (":_ppid:", pidstr);
+#endif
+    }
+
+
   };
   
   //      now, we shove the whole contents of the ENVIRON 
@@ -178,16 +220,16 @@ void crm_vht_init (int argc, char **argv)
 	    strcat (tempbuf, "\n");
 	  }
 	else
-	  untrappableerror ("The ENVIRONMENT variables don't fit into the "
+	  untrappableerror5 ("The ENVIRONMENT variables don't fit into the "
 			     "available space. \nThis is very broken.  Try "
 			     "a larger data window (with flag -w NNNNN), \nor "
 			     "drop the environment vars with "
-			     "the (with flag -e)", "");
+			     "the (with flag -e)", "", CRM_ENGINE_HERE);
 	while (environ[i][j] != '=') j++;
 	name = (char *) malloc ((sizeof (char)) * (j+200));
         if (!name)
-          untrappableerror("Couldn't malloc :_env_ space."  
-			   "Can't continue.\n","");
+          untrappableerror5("Couldn't malloc :_env_ space."  
+			   "Can't continue.\n","", CRM_ENGINE_HERE);
 	strcpy (name, ":_env_");
 	memmove (&(name[strlen(name)]), &(environ[i][0]), j);
 	name[j+6] = '\000';
@@ -317,9 +359,11 @@ void crm_vht_init (int argc, char **argv)
 }
 //           routine to put a variable into the temporary (tdw)
 //           buffer.  names and values end up interleaved
-//           sequentially, separated by newlines
+//           sequentially, separated by newlines.  TDW really should have
+//           been called the idw (Isolated Data Window) but it's too 
+//           late to fix it now.
 //
-
+//     
 void crm_set_temp_nvar (char *varname, char *value, long vallen)
 {
   long namestart, namelen;
@@ -332,16 +376,26 @@ void crm_set_temp_nvar (char *varname, char *value, long vallen)
     fprintf (stderr, "  setting temp-area variable %s to value %s\n",
 	     varname, value);
 
-  if (0 == crm_nextword (varname,strlen (varname), 0, &vnidx, &vnlen) )
+  i = crm_nextword (varname,strlen (varname), 0, &vnidx, &vnlen);
+  if ( i == 0)
     {
-      nonfatalerror ("Somehow, you are assigning a value to a variable with",
-     "an unprintable name.  I'll permit it for now, but your program is probably broken.");
+      nonfatalerror5 ("Somehow, you are assigning a value to a variable with",
+		     "an unprintable name.  I'll permit it for now, but"
+		      "your program is probably broken.", CRM_ENGINE_HERE);
     };
-      
+  
   if ( (strlen (varname) + vallen + tdw->nchars + 64) > data_window_size)
     {
-      fatalerror ("This program has overflowed the ISOLATEd data area with a variable that's just too big.  The bad variable was named: ", varname);
-      exit (EXIT_FAILURE);
+      fatalerror5 ("This program has overflowed the ISOLATEd data "
+		  "area with a variable that's just too big.  "
+		  "The bad variable was named: ",
+		   varname, CRM_ENGINE_HERE);
+      if (engine_exit_base != 0)
+	{
+	  exit (engine_exit_base + 22);
+	}
+      else
+	exit (EXIT_FAILURE);
     };
 
   //       check- is this the first time we've seen this variable?  Or
@@ -351,17 +405,18 @@ void crm_set_temp_nvar (char *varname, char *value, long vallen)
     {
       //  never assigned this variable before, so we stick it in the 
       //  tdr window. 
-      //       do the name first.  Start on a newline.
-      tdw->filetext[tdw->nchars] = '\n';
+      //
+      //       do the name first.  Start with a newline.
+      //  GROT GROT GROT
+      tdw->filetext[tdw->nchars] = 'X';
       tdw->nchars++;      
       namestart = tdw->nchars;
       namelen = vnlen;
-      // strcat (tdw->filetext, varname);
       memmove (&(tdw->filetext[tdw->nchars]), &(varname[vnidx]), namelen);
       tdw->nchars = tdw->nchars + namelen;
       //   
-      //        and add a separater to prevent the varname from sharing an edge
-      //        with the var value.
+      //        and add a separator to prevent the varname from sharing 
+      //        an endpoint offset with the var value.
       tdw->filetext[tdw->nchars] = '=';
       tdw->nchars++;
       //       
@@ -369,17 +424,21 @@ void crm_set_temp_nvar (char *varname, char *value, long vallen)
       valstart = tdw->nchars;
       memmove (&tdw->filetext [tdw->nchars], value, vallen);
       tdw->nchars = tdw->nchars + vallen;
-      //   
       //
+      //       add a separator again, so we don't get strings with overlapped
+      //       ranges into the var hash table
+      tdw->filetext[tdw->nchars] = 'Y';
+      tdw->nchars++;
+      //   
       //        and put a NUL at the end of the tdw, so debuggers won't get
       //       all bent out of shape.
       tdw->filetext[tdw->nchars] = '\000';
-	
+      //	
       //      now, we whack the actual VHT.
       crm_setvar (NULL, 0, 
 		  tdw->filetext, namestart, namelen,
 		  tdw->filetext, valstart, vallen, 
-		  0);
+		  0, 0);
       //     that's it.
 
     }
@@ -455,13 +514,16 @@ void crm_set_windowed_nvar ( char *varname,
       namelen = varlen;
       memmove (&tdw->filetext[namestart], varname, varlen);
       tdw->nchars = tdw->nchars + namelen;
+      //
+      //     put in an "&" separator
       tdw->filetext[tdw->nchars] = '&';
       tdw->nchars++;
+      //
       //      now, we whack the actual VHT.
       crm_setvar (NULL, 0, 
 		  tdw->filetext, namestart, namelen,
 		  valtext, start, len, 
-		  stmtnum);
+		  stmtnum, 0);
       //     that's it.
     }
   else
@@ -473,14 +535,6 @@ void crm_set_windowed_nvar ( char *varname,
       //     memory", and now inaccessible.
       //
       {
-	char *oldtext;
-	long oldstart, oldlen, oldend;
-	//     remember old values in case we have to reclaim this storage.
-	//   
-	oldtext = vht[i]->valtxt;
-	oldstart = vht[i]->vstart;
-	oldlen = vht[i]->vlen;
-	oldend = oldstart + oldlen;
 
 	//     move the text/start/len values around to accomodate the new
 	//     value.
@@ -490,7 +544,7 @@ void crm_set_windowed_nvar ( char *varname,
 	crm_setvar (NULL, 0,
 		    vht[i]->nametxt, vht[i]->nstart, vht[i]->nlen,
 		    valtext, start, len,
-		    stmtnum);
+		    stmtnum, 0);
 	
 	//       Do we need to repair the leaked memory?  Only necessary if the
 	//       old text was in the tdw area; this is harmless if the area
@@ -512,14 +566,28 @@ void crm_set_windowed_nvar ( char *varname,
     };
 }
 
-
+//#define RECLAIM_ALL_EVERY_TIME 1
+#ifdef RECLAIM_ALL_EVERY_TIME
 //
 //    How we compress out an area that might no longer be in use.
+static long crm_recursive_compress_tdw_section 
+(char *oldtext, long oldstart, long oldend);
 
 long crm_compress_tdw_section (char *oldtext, long oldstart, long oldend)
 {
+  //   let's court death, and do a FULL compress.
+  return (crm_recursive_compress_tdw_section 
+	  (tdw->filetext, 0, tdw->nchars + 1));
+}
+long crm_recursive_compress_tdw_section 
+(char *oldtext, long oldstart, long oldend)
+#else
+long crm_compress_tdw_section (char *oldtext, long oldstart, long oldend)
+#endif
+{
+  
   //  The algorithm basically checks to see if there is any region of
-  //  the old tdw space that is not currently used by another var.
+  //  the given tdw space that is not currently used by another var.
   //  All such regions are reclaimed with a slice-n-splice.  We return
   //  the number of reclaimed characters.
   //
@@ -527,22 +595,47 @@ long crm_compress_tdw_section (char *oldtext, long oldstart, long oldend)
   //  unused "to be killed" region.  It checks each member of the VHT
   //  in the TDW.  If the region overlaps, don't kill the overlapping
   //  part of the region.  If at any time the region length goes to 0,
-  //  we know that there's no region left to kill.
+  //  we know that there's no region left to kill.  (Option- if the
+  //  gap is less than MAX_RECLAIMER_GAP chars, we don't bother moving
+  //  it; we retain it as buffer.  This minimizes thrashing 
   //  
-  //  Note that the "oldend" and "newend" vars are NON-inclusive, they
-  //  index the first NON-involved character (oldstart and newstart index
-  //  "involved" characters, that we _do_ include in our strings)
+  //  NOTE that the END VALUES ONLY "oldend" and "newend" vars are
+  //  NON-inclusive, they index the first NON-involved character
+  //  (oldstart and newstart index "involved" characters, that we _do_
+  //  include in our strings)
+  //
+  //  BIG ISSUE: As coded, this routine needs to leave a _buffer_ of at
+  //  least one UNUSED character between each used (but isolated) string
+  //  area.  Knowing when to get rid of extra copies of this character
+  //  has been a big hassle.  Right now there may be a small leak here
+  //  so if you can find it, please let me know!   Note that any fix that
+  //  does not keep two adjacent isolated regions from merging (including 
+  //  when the first or second becomes a zero-length string!) will get 
+  //  the submittor a gentle smile and a pointer to this very comment.
+  //  (the reason being that prior code that did not leave a buffer
+  //  exhibited the property that if A and B were isolated but adjacent,
+  //  and then A shrank to 0 length, then B would share the same start
+  //  point, and an alteration to A would then *also* insert at the start
+  //  point of B, causing A and B to become NONisolated and space-sharing.
+  //  That said- enjoy the bug hunt.  :)
 
-  long j, newstart, newend; 
+  long j, newstart, newend, reclaimed;
+
+  j = newstart = newend = reclaimed = 0;
 
   //  return (0);
 
+  j = newstart = newend = reclaimed = 0;
   if (internal_trace)
-    fprintf (stderr, " [ Compressing isolated data.  Length %ld chars, start %ld, len %ld ]\n", tdw->nchars, oldstart, oldend - oldstart);
+    fprintf (stderr, " [ Compressing isolated data.  Length %ld chars, "
+	     "start %ld, len %ld ]\n", 
+	     tdw->nchars, 
+	     oldstart, 
+	     oldend - oldstart);
 
   //      If oldstart >= oldend, then there's no compression to be done.
   //
-  if (oldstart >= oldend && 0)
+  if (oldstart >= oldend )
     {
       if (internal_trace)
 	fprintf (stderr, " [ Zero-length compression string... don't do this! ]\n");
@@ -551,22 +644,29 @@ long crm_compress_tdw_section (char *oldtext, long oldstart, long oldend)
 
   if (oldtext != tdw->filetext) 
     {
-      nonfatalerror (" Request to compress non-TDW data.  This is bogus. ",
-		     " Please file a bug report");
+      fatalerror5 (" Request to compress non-TDW data.  This is bogus. ",
+		  " Please file a bug report", CRM_ENGINE_HERE);
       return ( 0 );
     };
 
+  //    Look one character further to before and after;
+  //if (oldstart > 3)  oldstart --;
+  //if (oldend < data_window_size - 1) oldend ++;
+
   for (j = 0; j < vht_size; j++)
     {
       if (vht[j]    // is this slot in use?
 	  && vht[j]->valtxt == tdw->filetext
-	  && vht[j]->vstart != 0 )              // this excludes the tdw itself
+	  //   Note that being part of :_iso: does NOT exclude from reclamation
+	  &&  0 != strncmp (&vht[j]->nametxt[vht[j]->nstart], ":_iso:", 6 ))
 	{
-	  //    for convenience, some copies of vars
-	  newstart = vht[j]->vstart;
-	  newend = newstart + vht[j]->vlen;
-	    
-	  //    Possible cases:
+	  //    for convenience, we get nice short names:
+	  newstart = vht[j]->vstart - 1;
+	  newend = newstart + vht[j]->vlen + 2;
+	  //  leave some space no matter what...
+	  if (newend < newstart + 2) newend = newstart + 2;
+    
+	  //    6 Possible cases:
 	  //      dead zone entirely before current var  
 	  //      dead zone entirely after current var
 	  //      dead zone entirely inside current var
@@ -575,15 +675,149 @@ long crm_compress_tdw_section (char *oldtext, long oldstart, long oldend)
 	  //      dead zone split by current var
 	  //
 
-	  //      dead zone entirely before current var  
+	  //      1: dead zone entirely before current var  
 	  //  
 	  //       <os------------oe>
 	  //                           <ns--------ne>
 	  //
-	  if ( oldend < newstart)
+	  if ( oldend <= newstart)
 	    {
 	      //   nothing to be done here - not overlapping 
-	      goto end_of_tests;
+	      goto end_of_vstring_tests;
+	    };
+
+	  //      2: dead zone entirely after current var
+	  //  
+	  //                        <os------------oe>
+	  //       <ns--------ne>
+	  //
+	  if ( newend <= oldstart )
+	    {
+	      //   nothing to be done here - not overlapping
+	      goto end_of_vstring_tests;
+	    };
+
+	  //   If we get this far, the dead zone in some way overlaps with
+	  //   our current variable.
+
+	  //      3: dead zone entirely inside a currently live var
+	  //
+	  //                 <os-------oe>
+	  //              <ns----------------ne>
+	  //
+	  //      So we terminate this procedure (nothing can be reclaimed)
+	  //
+	  if (oldstart >= newstart && oldend <= newend)
+	    {
+	      //   the dead zone is inside a non-dead var, so 
+	      //   we can terminate our search right now.
+	      if ( internal_trace)
+		fprintf (stderr, " [ Compression not needed after all. ]\n");
+	      return ( 0 );
+	    };
+
+	  //      4: dead zone overlaps front of current var; we trim the 
+	  //      dead zone to not include the current var.
+	  //  
+	  //       <os------------oe>
+	  //              <ns--------ne>
+	  //
+	  if ( oldstart < newstart && oldend <= newend )  
+	    {
+	      //   The dead zone should not include the part that's 
+	      //   also new variable.  So, we clip out the part
+	      //   that's still active.
+	      if ( internal_trace)
+		fprintf (stderr, " [ Trimming tail off of compression. ]\n");
+	      //
+	      //     newstart is a "good" char, but since oldend is
+	      //     noninclusive, this is right.
+	      oldend = newstart; 
+	      goto end_of_vstring_tests;
+	    };
+	  
+	  //      5: dead zone overlaps back of current var; trim the front off
+	  //      the dead zone.
+	  //
+	  //                   <os------------oe>
+	  //              <ns--------ne>
+	  //
+	  if (newstart <= oldstart && newend <= oldend)
+	    {
+	      if (internal_trace)
+		fprintf (stderr, " [ Trimming head off of compression. ]\n");
+	      //
+	      //     Newend is the first char that ISN'T in the var, so this
+	      //     is correct.
+	      oldstart = newend ;
+	      goto end_of_vstring_tests;
+	    };
+	  //      6: dead zone split by current var - the dead zone is actually
+	  //      split into two distinct pieces.  In this case, we need to
+	  //      recurse on the two pieces.
+	  //          
+	  //         <os--------------------oe>
+	  //               <ns--------ne>
+	  //
+	  if ( oldstart <= newstart && newend <= oldend )
+	    {
+	      if (internal_trace)
+		{
+		  fprintf (stderr, " [ Compression split ]\n");
+		  fprintf (stderr, " [ First part will be %ld to %ld .]\n",
+			   oldstart, newstart);
+		  fprintf (stderr, " [ Second part will be %ld to %ld .]\n",
+			   newend, oldend);
+		};
+	      //
+	      //      Tricky bit here - we have to do the aft (ne-oe
+	      //      section) first, so we don't move the os-ns
+	      //      section offsets.
+	      //
+
+		      //    was newend - 1, but should be same as case 3
+		      //   above (dead zone overlaps tail)
+#ifdef RECLAIM_ALL_EVERY_TIME
+	      reclaimed = crm_recursive_compress_tdw_section (oldtext, newend, oldend);
+	      reclaimed +=crm_recursive_compress_tdw_section(oldtext, oldstart, newstart);
+#else
+	      reclaimed = crm_compress_tdw_section (oldtext, newend, oldend);
+	      reclaimed +=crm_compress_tdw_section(oldtext, oldstart, newstart);
+#endif
+	      //   Return here instead of executing common slice-and-splice 
+	      //   tail, because each of our recursive children will do
+	      //   that for us.
+	      return (reclaimed);
+	    }
+	}
+      //       and the semicolon to keep some compilers happy
+    end_of_vstring_tests: ;
+      //    Now, repeat with the name string - all name strings are protected
+      if (vht[j]    
+	  && vht[j]->nametxt == tdw->filetext) 
+	{
+	  newstart = vht[j]->nstart - 1 ;
+	  newend = newstart + vht[j]->nlen + 2;
+	  //  leave some space no matter what...
+	  if (newend < newstart + 4) newend = newstart + 2;
+	  
+	  //    Possible cases:
+	  //      dead zone entirely before current var  
+	  //      dead zone entirely after current var
+	  //      dead zone entirely inside current var
+	  //      dead zone overlaps front of current var
+	  //      dead zone overlaps back of current var
+	  //      dead zone split by current var
+	  //
+	  //      dead zone entirely before current var  
+	  //  
+	  //       <os------------oe>
+	  //                           <ns--------ne>
+	  // OK
+	  if ( oldend <= newstart)
+	    {
+	      //   nothing to be done here - not overlapping 
+	      goto end_of_nstring_tests;
 	    };
 
 	  //      dead zone entirely after current var
@@ -591,10 +825,10 @@ long crm_compress_tdw_section (char *oldtext, long oldstart, long oldend)
 	  //                        <os------------oe>
 	  //       <ns--------ne>
 	  //
-	  if ( newend < oldstart )
+	  if ( newend <= oldstart )
 	    {
 	      //   nothing to be done here - not overlapping
-	      goto end_of_tests;
+	      goto end_of_nstring_tests;
 	    };
 
 	  //   If we get this far, the dead zone in some way overlaps with
@@ -621,8 +855,8 @@ long crm_compress_tdw_section (char *oldtext, long oldstart, long oldend)
 	  //  
 	  //       <os------------oe>
 	  //              <ns--------ne>
-	  //
-	  if ( oldstart <= newstart && oldend <= newend )  
+	  // 
+	  if ( oldstart < newstart && oldend <= newend )  
 	    {
 	      //   The dead zone should not include the part that's 
 	      //   also new variable.  So, we clip out the part
@@ -633,7 +867,7 @@ long crm_compress_tdw_section (char *oldtext, long oldstart, long oldend)
 	      //     newstart is a "good" char, but since oldend is
 	      //     noninclusive, this is right.
 	      oldend = newstart; 
-	      goto end_of_tests;
+	      goto end_of_nstring_tests;
 	    };
 	  
 	  //      dead zone overlaps back of current var; trim the front off
@@ -650,7 +884,7 @@ long crm_compress_tdw_section (char *oldtext, long oldstart, long oldend)
 	      //     Newend is the first char that ISN'T in the var, so this
 	      //     is correct.
 	      oldstart = newend ;
-	      goto end_of_tests;
+	      goto end_of_nstring_tests;
 	    };
 	  //      dead zone split by current var - the dead zone is actually
 	  //      split into two distinct pieces.  In this case, we need to
@@ -667,21 +901,30 @@ long crm_compress_tdw_section (char *oldtext, long oldstart, long oldend)
 		  fprintf (stderr, " [ First part will be %ld to %ld .]\n",
 			   oldstart, newstart);
 		  fprintf (stderr, " [ Second part will be %ld to %ld .]\n",
-			   newend - 1, oldend);
+			   newend, oldend);
 		};
 	      //
 	      //      Tricky bit here - we have to do the aft (ne-oe
 	      //      section) first, so we don't move the os-ns
 	      //      section offsets.
 	      //
-	      return (
+
 		      //    was newend - 1, but should be same as case 3
 		      //   above (dead zone overlaps tail)
-		      crm_compress_tdw_section (oldtext, newend, oldend)
-		      + crm_compress_tdw_section(oldtext, oldstart, newstart));
+#ifdef RECLAIM_ALL_EVERY_TIME
+	      reclaimed = crm_recursive_compress_tdw_section (oldtext, newend, oldend);
+	      reclaimed +=crm_recursive_compress_tdw_section (oldtext, oldstart, newstart);
+#else
+	      reclaimed = crm_compress_tdw_section (oldtext, newend, oldend);
+	      reclaimed +=crm_compress_tdw_section (oldtext, oldstart, newstart);
+#endif
+	      //   Return here instead of executing common slice-and-splice 
+	      //   tail, because each of our recursive children will do
+	      //   that for us.
+	      return (reclaimed);
 	    }
 	  //       and the semicolon to keep some compilers happy
-	end_of_tests: ;
+	end_of_nstring_tests: ;
 	};
     };
   //
@@ -693,13 +936,21 @@ long crm_compress_tdw_section (char *oldtext, long oldstart, long oldend)
   //
   {
     long cutlen;
-    cutlen =  -(oldend - (oldstart));
+    // cutlen is supposed to be negative for compress
+    cutlen = oldstart - oldend - 1;
     if (cutlen > 0)
-      nonfatalerror ("Internal cut-length error in isolated var reclamation.",
-		     "  Please file a bug report");
+      fatalerror5 ("Internal cut-length error in isolated var reclamation.",
+		  "  Please file a bug report", CRM_ENGINE_HERE);
+
+    //    Future Enhancement - dead zones of some small size should be
+    //    allowed to stay.  This would speed up WINDOW a lot. (but we
+    //    would need to expand the range of oldstart and oldend to
+    //    actually reclaim those areas if storage really ran low.
+    //    Maybe this should be compile-time or command-line parameter?)
+
     if (cutlen < 0)
       {
-	if (internal_trace)
+        if (internal_trace)
 	  {
 	    fprintf (stderr, " [ compression slice-splice at %ld for %ld chars. ]\n",
 		     oldstart, cutlen);
@@ -727,16 +978,18 @@ void crm_destructive_alter_nvariable (char *varname, long varlen,
   long i;
   long vhtindex, oldlen, delta;
 
-  //      get the first variable name and verify it exists.
+  //      get the first variable name and verify it exists.  
+  //     GROT GROT GROT this should use nextword!!!  
   i = 0;
   while (varname[i] < 0x021 && i < varlen) i++;
   vhtindex = crm_vht_lookup (vht, &(varname[i]), varlen);
   if (vht[vhtindex] == NULL)
     {
-      nonfatalerror(" Attempt to alter the value of a nonexistent "
-		    "variable, so I'm creating an ISOLATED variable.  "
-		    "I hope that's OK.  The nonexistent variable is: ",
-		    &(varname[i]));
+      // IGNORE FOR NOW
+      nonfatalerror5 (" Attempt to alter the value of a nonexistent "
+      		    "variable, so I'm creating an ISOLATED variable.  "
+      		    "I hope that's OK.  The nonexistent variable is: ",
+      		    &(varname[i]), CRM_ENGINE_HERE);
       crm_set_temp_var (&varname[i], "");
     };
   
@@ -748,16 +1001,29 @@ void crm_destructive_alter_nvariable (char *varname, long varlen,
     mdw = tdw;
   if (cdw->filetext == vht[vhtindex]->valtxt)
     mdw = cdw;
+
+  //   GROT GROT GROT  get rid of this if we go to MAPped file vars.
   if (mdw == NULL)
     {
-      fatalerror (" Bogus text bloc containing variable ", varname);  
+      fatalerror5 (" Bogus text bloc containing variable : ", 
+		   varname, CRM_ENGINE_HERE);  
       goto bailout;
     };
+
   //
   if (user_trace)  // major debug
     {
-      fprintf (stderr, "\n     surgery on the var %s, ", varname);
-      fprintf (stderr, "new value is: \n***%s***\n", newstr);
+      long i;
+      // fprintf (stderr, "\n     surgery on the var %s\n ", varname);
+      fprintf (stderr, " surgery on the var >");
+      for (i = 0; i < varlen; i++ )
+	fprintf (stderr, "%c", varname[i]);
+      fprintf (stderr, "<\n");
+      //fprintf (stderr, "new value is: \n***%s***\n", newstr);
+      fprintf (stderr, " new value is ***>");
+      for (i = 0; i < newlen; i++ )
+	fprintf (stderr, "%c", newstr[i]);
+      fprintf (stderr, "<***\n");
     }
   //     slice and splice the mdw text area, to make the right amount of 
   //     space...
@@ -803,8 +1069,9 @@ void crm_slice_and_splice_window ( CSL_CELL *mdw, long where, long delta)
 
   if (delta + mdw->nchars > data_window_size - 10)
     {
-      fatalerror (" Data window trying to get too long.", 
-		" Try increasing the data window maximum size.");
+      fatalerror5 (" Data window trying to get too long.", 
+		   " Try increasing the data window maximum size.",
+		   CRM_ENGINE_HERE);
       goto bailout;
     };
 
@@ -846,7 +1113,7 @@ void crm_slice_and_splice_window ( CSL_CELL *mdw, long where, long delta)
 	     (long int) tailsrc, (long int) taildest, taillen);
 
   //     and move the actual data
-  memmove ( taildest, tailsrc, taillen + 1 );
+  if (taillen + 1 > 0) memmove ( taildest, tailsrc, taillen + 1 );
 
   //     update the length of the window as well.
   mdw->nchars = mdw->nchars + delta;
@@ -876,11 +1143,12 @@ void crm_slice_and_splice_window ( CSL_CELL *mdw, long where, long delta)
       long odws, i;
       odws = data_window_size;
       data_window_size = 4 * data_window_size;
-      nonfatalerror (" Data window trying to get too long.", 
-  		   " increasing data window... ");
+      nonfatalerror5 (" Data window trying to get too long.", 
+		     " increasing data window... ", CRM_ENGINE_HERE);
       ndw = (char *) malloc ( data_window_size);
       if (!ndw)
-          untrappableerror("Couldn't malloc ndw.  This is bad too.\n","");
+	untrappableerror5("Couldn't malloc ndw.  This is bad too.\n","",
+			  CRM_ENGINE_HERE);
 
       //  now copy the old data window into the new one
       memmove (ndw, mdw->filetext, odws);
@@ -914,13 +1182,49 @@ long crm_vht_lookup (VHT_CELL **vht, char *vname, long vlen)
   unsigned long hc;
   unsigned long i, j, k;
   int done;
-  unsigned long vsidx;
-  unsigned long vslen;
+  long vsidx;
+  long vslen;
 
   j = 0;  // just so J is used.
 
-  crm_nextword ( vname, vlen, 0, &vsidx, &vslen);
+  //   Consistency scan - look for those null varnames!  Do this every
+  //   time!
+  if (1)
+    {
+      long i, j;
+      long corrupted;
+      for (i = 0; i < vht_size; i++)
+	{
+	  corrupted = 0;
+	  if (vht[i] != NULL && vht[i]->nlen < 2)
+	    fprintf (stderr, "Short length %ld ", i);
+	  if (vht[i] !=NULL && vht[i]->nlen > 1)
+	    {
+	      if (vht[i]->nametxt[vht[i]->nstart] != ':')
+		{
+		  fprintf (stderr, "Ztart corrupted ");
+		  corrupted = 1;
+		};
+	      if (vht[i]->nametxt[vht[i]->nstart + vht[i]->nlen - 1] != ':')
+		{
+		  fprintf (stderr, "Zend corrupted ");
+		  corrupted = 1;
+		};
+	      if (corrupted)
+		{
+		  fprintf (stderr, " i %ld len %ld name = -",
+			   i, vht[i]->nlen );
+		  for (j = 0; j < vht[i]->nlen; j++)
+		    fprintf (stderr, "%c", 
+			     vht[i]->nametxt[vht[i]->nstart + j]);
+		  fprintf (stderr, "- ");
+		}
+	    };
+	}
+    };
+  
 
+  crm_nextword ( vname, vlen, 0, &vsidx, &vslen);
   if (internal_trace)
     {
       fprintf (stderr, " variable len %ld, name is -", vslen);
@@ -957,13 +1261,12 @@ long crm_vht_lookup (VHT_CELL **vht, char *vname, long vlen)
 	      for (ic = 0; ic < vlen; ic++)
 		fprintf (stderr, "%c", vname[ic]);
 	      fprintf (stderr, "(len %ld) not at %ld (empty)\n", vlen, i);
+	      fprintf (stderr, "Returning the index where it belonged.\n");
 	    };
 	  return (i);
 	};
       
       //  there's something here - is it what we have been seeking
-
-      //      if (strncmp (&((vht[i]->nametxt)[vht[i]->nstart]),
       if ( vlen == vht[i]->nlen &&
 	   memcmp  (&((vht[i]->nametxt)[vht[i]->nstart]),
 		   vname,
@@ -1012,8 +1315,32 @@ long crm_vht_lookup (VHT_CELL **vht, char *vname, long vlen)
       //   get close to overflow.
       if (i == (hc - 1))
 	{
-	  fatalerror (" hash table overflow while looking for variable: " ,
-		      "");
+	  static char badvarname [MAX_VARNAME];
+	  strncpy (badvarname, &vname[vsidx], vslen);
+	  badvarname[vslen+1] = 0;
+	  {
+	    long index;
+	    fprintf (stderr, "Variable Hash Table Dump\n");
+	    for (index = 0; index < vht_size; index++)
+	      {
+		int ic;
+		fprintf (stderr, "  var '");
+		for (ic = 0; ic < vht[index]->nlen; ic++)
+		  fprintf (stderr, "%c", 
+			   (vht[index]->nametxt) [ic+vht[index]->nstart] );
+		fprintf (stderr, "'[%ld] found at %ld (",
+			 vht[index]->nlen,  index);
+		if (vht[index]->valtxt == cdw->filetext) 
+		  { fprintf (stderr, "(main)"); }
+		else
+		  { fprintf (stderr, "(isol)"); };
+		fprintf (stderr, " s: %ld, l:%ld)\n",
+		     vht[index]->vstart, vht[index]->vlen);
+		; }
+	  };
+	  fatalerror5 (" Variable hash table overflow while looking "
+		      "for variable: " ,
+		       badvarname, CRM_ENGINE_HERE);
 	  done = 1;
 	  return (0);
 	};
@@ -1038,7 +1365,8 @@ void crm_setvar (
 		 char *valtxt,  
 		 long vstart,   
 		 long vlen,     
-		 long linenumber
+		 long linenumber,
+		 long lazy_redirects
 		 )
 {
   int i, j;     // some indices to bang on
@@ -1055,7 +1383,7 @@ void crm_setvar (
       //  allocate a fresh, empty VHT cell
       vht[i] = (VHT_CELL *) malloc (sizeof (VHT_CELL));
       if (!vht[i])
-	untrappableerror("Couldn't malloc space for VHT cell.  We need VHT cells for variables.  We can't continue.","");
+	untrappableerror5("Couldn't malloc space for VHT cell.  We need VHT cells for variables.  We can't continue.","", CRM_ENGINE_HERE);
       
       //  fill in the name info data
       vht[i]->filename = filename;
@@ -1063,10 +1391,10 @@ void crm_setvar (
       vht[i]->nametxt = nametxt;
       vht[i]->nstart = nstart;
       vht[i]->nlen = nlen;
-      vht[i]->mstart = vstart;
-      vht[i]->mlen = 0;
       vht[i]->vstart = 0 ;
-      vht[i]->vlen = 0;
+      vht[i]->vlen = 0;\
+      vht[i]->lazy_redirects = lazy_redirects;
+
       //  and now that the slot has proper initial information,
       //  we can use the same code as is used in an update to do 
       //  the initial setting of values.  This is good because 
@@ -1075,7 +1403,7 @@ void crm_setvar (
     }
   else
     {
-
+      //   The cell is already here.  :)
     };
   //   Either way, the cell is now here, so we can set the value.
   //
@@ -1085,6 +1413,7 @@ void crm_setvar (
   vht[i]->mstart = vstart;
   vht[i]->mlen = 0;
   vht[i]->linenumber = linenumber;
+  vht[i]->lazy_redirects = lazy_redirects;
 
   if(internal_trace)
     {
@@ -1104,6 +1433,8 @@ void crm_setvar (
       
       fprintf (stderr, "- (start %ld, length %ld)", 
 	       vht[i]->vstart, vht[i]->vlen); 
+      
+      fprintf (stderr, "and %ld lazy redirects", vht[i]->lazy_redirects);
 
       fprintf (stderr, "\n");
     };
@@ -1158,13 +1489,13 @@ long crm_lookupvarline (VHT_CELL **vht, char *text, long start, long len)
 
 void crm_updatecaptures (char *text, long loc, long delta)
 {
-  int vht_index;
+  long vht_index;
   long i;
-  long ostart, oend;
-  long nstart, nend;
+  long ostart = 0, oend = 0 ;
+  long nstart = 0, nend = 0 ;
 
   if (internal_trace)
-    fprintf (stderr, "\n    updating captured values start %ld len %ld\n",
+    fprintf (stderr, "\n    updating captured values start %ld len %ld \n",
 	     loc, delta);
 
   //        check each VHT entry for a need to relocate
@@ -1174,7 +1505,7 @@ void crm_updatecaptures (char *text, long loc, long delta)
       if (vht[vht_index] != NULL)
 	{
 	  if (vht[vht_index]->valtxt == text)
-	    {      
+	    {  // start of valtext block check      
 	      // value text area
 	      if (internal_trace > 1)
 		{
@@ -1191,82 +1522,68 @@ void crm_updatecaptures (char *text, long loc, long delta)
 			   vht[vht_index]->nstart + vht[vht_index]->nlen
 			   );
 		};
-	      ostart = vht[vht_index]->vstart;
-	      oend = ostart + vht[vht_index]->vlen;
+	      ostart = nstart = vht[vht_index]->vstart;
+	      oend = nstart = ostart + vht[vht_index]->vlen;
 	      nstart = crm_mangle_offset (ostart, loc, delta, 0);
 	      nend = crm_mangle_offset (oend, loc, delta, 1);
-	      if (ostart == nstart && oend == nend )
-		{
-		  //   no change
-		}
-	      else
-		{
-		  if (internal_trace)
-		    fprintf (stderr, "\n   index %d vstart/vlen upd: %ld, %ld  ",
-			     vht_index, 
-			     vht[vht_index]->vstart, vht[vht_index]->vlen);
-		  vht[vht_index]->vstart = nstart;
-		  vht[vht_index]->vlen = nend - nstart;
-		  if (internal_trace) 
-		    fprintf (stderr, "to %ld, %ld.\n",
-			     vht[vht_index]->vstart,
-			     vht[vht_index]->vlen);
-		};
+	      if (internal_trace)
+		fprintf (stderr, "\n   index %ld vstart/vlen upd: %ld, %ld  ",
+			 vht_index, 
+			 vht[vht_index]->vstart, vht[vht_index]->vlen);
+	      vht[vht_index]->vstart = nstart;
+	      vht[vht_index]->vlen = nend - nstart;
+	      if (internal_trace) 
+		fprintf (stderr, "to %ld, %ld.\n",
+			 vht[vht_index]->vstart,
+			 vht[vht_index]->vlen);
 	      //   
 	      //        And do the same for mstart/mlen (match start/length)
 	      ostart = vht[vht_index]->mstart;
 	      oend = ostart + vht[vht_index]->mlen;
 	      nstart = crm_mangle_offset (ostart, loc, delta, 0);
 	      nend = crm_mangle_offset (oend, loc, delta, 1);
-	      if (ostart == nstart 
-		  && oend == nend )
-		{
-		  //   no change
-		}
-	      else
-		{
-		  if (internal_trace) 
-		    fprintf (stderr, "\n     index %d mstart/mlen upd: %ld, %ld  ",
-			     vht_index, 
-			     vht[vht_index]->mstart, vht[vht_index]->mlen);
-		  vht[vht_index]->mstart = nstart;
-		  vht[vht_index]->mlen = nend - nstart;
-		  if (internal_trace) 
-		    fprintf (stderr, "to %ld, %ld.\n",
-			     vht[vht_index]->mstart,
-			     vht[vht_index]->mlen);
-		};
-
+	      if (internal_trace) 
+		fprintf (stderr, "\n index %ld mstart/mlen upd: %ld, %ld  ",
+			 vht_index, 
+			 vht[vht_index]->mstart, vht[vht_index]->mlen);
+	      vht[vht_index]->mstart = nstart;
+	      vht[vht_index]->mlen = nend - nstart;
+	      if (internal_trace) 
+		fprintf (stderr, "to %ld, %ld.\n",
+			 vht[vht_index]->mstart,
+			 vht[vht_index]->mlen);
 	    };
 	  //      Don't forget entries that may be varNAMES, not just
 	  //      var values!
 	  if (vht[vht_index]->nametxt == text)
 	    {
+	      long orig_len;
 	      //
 	      //    Same thing here...
 	      //
-	      ostart = vht[vht_index]->nstart;
-	      oend = ostart + vht[vht_index]->nlen;
+	      ostart = nstart = vht[vht_index]->nstart;
+	      orig_len = vht[vht_index]->nlen;
+	      oend = nend = ostart + orig_len;
+	      if (orig_len == 0) fprintf (stderr, "CRUD on %ld", vht_index);
 	      nstart = crm_mangle_offset (ostart, loc, delta, 0);
 	      nend = crm_mangle_offset (oend, loc, delta, 1);
-	      if (ostart == nstart && oend == nend )
-		{
-		  //   no change
-		}
-	      else
-		{
-		  if (internal_trace) 
-		    fprintf (stderr, 
-			     "\n      index %d nstart/nlen upd: %ld, %ld  ",
-			     vht_index, 
-			     vht[vht_index]->nstart, vht[vht_index]->nlen);
-		  vht[vht_index]->nstart = nstart;
-		  vht[vht_index]->nlen = nend - nstart;
-		  if (internal_trace) 
-		    fprintf (stderr, "to %ld, %ld.\n",
-			     vht[vht_index]->nstart,
-			     vht[vht_index]->nlen);
-		};
+	      if (oend - ostart != orig_len) 
+		fprintf (stderr, "Length change on %ld!  Was %ld, now %ld ",
+			 vht_index,
+			 orig_len,
+			 oend-ostart);
+
+	      if (internal_trace) 
+		fprintf (stderr, 
+			 "\n      index %ld nstart/nlen upd: %ld, %ld  ",
+			 vht_index, 
+			 vht[vht_index]->nstart, vht[vht_index]->nlen);
+	      vht[vht_index]->nstart = nstart;
+	      vht[vht_index]->nlen = nend - nstart;
+	      if (internal_trace) 
+		fprintf (stderr, "to %ld, %ld.\n",
+			 vht[vht_index]->nstart,
+			 vht[vht_index]->nlen);
 	    }
 	}
     }
@@ -1380,6 +1697,6 @@ long crm_mangle_offset ( long mark, long dot, long delta, long sl)
 int crm_buffer_gc ( CSL_CELL *zdw)
 {
   fprintf (stderr, "Sorry, GC is not yet implemented");
-  exit (1);
+  exit (EXIT_FAILURE);
   return (0);
 }
